import sys
import time

import numpy as np
import pandas as pd
import rasterio
from loguru import logger

from openamundsen import (
    conf,
    constants,
    errors,
    fileio,
    forcing,
    modules,
    statevars,
    surface,
    terrain,
    util,
)
from openamundsen import (
    meteo as oameteo,
)

from .landcover import LandCover


class OpenAmundsen:
    """
    Class encapsulating the required data and methods for a single
    openAMUNDSEN model run.

    Parameters
    ----------
    config : dict
        Model run configuration.

    Examples
    --------
    >>> model = oa.OpenAmundsen(oa.read_config('config.yml'))
    >>> model.initialize()
    >>> model.run()
    """

    def __init__(self, config):
        self.config = conf.parse_config(config)
        self.state = None
        self.dates = None
        self.date = None
        self.date_idx = None

    def initialize(self, meteo=None, meteo_callback=None):
        """
        Initialize the model according to the given configuration, i.e. read
        the required input raster files and meteorological input data,
        initialize the model grid and all required state variables, etc.
        """
        config = self.config

        self._require_soil = config.snow.model == "multilayer"
        self._require_energy_balance = config.snow.melt.method == "energy_balance"
        self._require_temperature_index = not self._require_energy_balance
        self._require_canopy = config.canopy.enabled
        self._require_evapotranspiration = config.evapotranspiration.enabled
        self._require_land_cover = self._require_canopy or self._require_evapotranspiration
        self._require_soil_texture = self._require_evapotranspiration
        self._require_interpolation = config.input_data.meteo.format in ("csv", "netcdf", "memory")
        self._require_snow_management = (
            conf.SNOW_MANAGEMENT_AVAILABLE and config.snow_management.enabled
        )
        self._require_glaciers = config.glaciers.enabled

        self.configure_logger()

        self._prepare_time_steps()
        self._initialize_grid()
        self._initialize_state_variable_management()

        if config.snow.model == "multilayer":
            self.snow = modules.snow.MultilayerSnowModel(self)
        elif config.snow.model == "cryolayers":
            self.snow = modules.snow.CryoLayerSnowModel(self)
        else:
            raise NotImplementedError

        if self._require_land_cover:
            self.land_cover = LandCover(self)

        if self._require_canopy:
            self.canopy = modules.canopy.CanopyModel(self)

        if self._require_evapotranspiration:
            self.evapotranspiration = modules.evapotranspiration.EvapotranspirationModel(self)

        if self._require_glaciers:
            logger.warning(
                "The glacier model is experimental and under active development. Use with caution."
            )
            self.glaciers = modules.glacier.GlacierModel(self)

        if config.meteo.interpolation.wind.method == "liston":
            self.state.base.add_variable(
                "scaled_curvature",
                "1",
                "Topographic curvature scaled to [-0.5, 0.5]",
                retain=True,
            )

        # Create snow redistribution factor state variables
        for precip_corr in config.meteo.precipitation_correction:
            if precip_corr["method"] == "srf":
                self.state.base.add_variable("srf", "1", "Snow redistribution factor", retain=True)
                break  # multiple SRFs are not allowed

        if self._require_snow_management:
            import openamundsen_snowmanagement

            self.snow_management = openamundsen_snowmanagement.SnowManagementModel(self)

        self._create_state_variables()

        self._read_input_data()

        if config.input_data.meteo.format in ("netcdf", "csv"):
            meteo = self._read_meteo_data()
        elif config.input_data.meteo.format == "memory":
            if meteo is None:
                raise errors.MeteoDataError(
                    "A meteo dataset must be passed to "
                    "OpenAmundsen.initialize() if the meteo input format "
                    'is set to "memory"'
                )

            if not forcing.is_valid_point_dataset(meteo, dates=self.dates):
                raise errors.MeteoDataError("Not a valid point forcing dataset")

            meteo = meteo.copy(deep=True)
        elif config.input_data.meteo.format == "callback":
            if meteo_callback is None:
                raise errors.MeteoDataError(
                    "meteo_callback must be passed to "
                    "OpenAmundsen.initialize() if the meteo input format "
                    'is set to "callback"'
                )
            elif not callable(meteo_callback):
                raise errors.MeteoDataError(
                    "meteo_callback must be callable (i.e., a function or method)"
                )

            self._meteo_callback = meteo_callback

            # Create a dummy point forcing dataset with 0 stations
            dummyds = forcing.make_empty_point_dataset(
                self.dates,
                "dummy",
                "dummy",
                np.nan,
                np.nan,
                np.nan,
            )
            meteo = forcing.combine_point_datasets([dummyds]).drop_isel(station=0)

        if (
            config.meteo.interpolation.cloudiness.method == "prescribed"
            and "cloud_fraction" not in meteo
        ):
            raise errors.MeteoDataError(
                'Cloud cover data must be provided for cloudiness method "prescribed"'
            )

        self.meteo = forcing.prepare_point_coordinates(meteo, self.grid, self.config.crs)
        oameteo.correct_station_precipitation(self)

        self._has_wind_gusts = "wind_speed_gust" in self.meteo

        # Extend ROI with the station positions
        if config.extend_roi_with_stations:
            pos_outside_roi_stations = self.meteo.within_grid_extent & ~self.meteo.within_roi
            any_outside_roi_stations = np.any(pos_outside_roi_stations)
            if any_outside_roi_stations:
                self.meteo.within_roi[pos_outside_roi_stations] = True
                outside_roi_rows = self.meteo.row[pos_outside_roi_stations].values
                outside_roi_cols = self.meteo.col[pos_outside_roi_stations].values
                self.grid.roi[outside_roi_rows, outside_roi_cols] = True
                self.grid.prepare_roi_coordinates()

        self._calculate_terrain_parameters()

        config.results_dir.mkdir(parents=True, exist_ok=True)  # create results directory if necessary # noqa: E501 # fmt: skip
        self._initialize_point_outputs()
        self._initialize_gridded_outputs()

        self._initialize_state_variables()

        if self.config.liveview.enabled:
            logger.info("Creating live view window")
            from openamundsen import liveview

            lv = liveview.LiveView(self.config.liveview, self.state, self.grid.roi)
            lv.create_window()
            self.liveview = lv

    def run(self):
        """
        Start the model run. Before calling this method, the model must be
        properly initialized by calling `initialize()`.
        """
        logger.info("Starting model run")
        start_time = time.time()

        for _ in range(len(self.dates)):
            self.run_single()

        time_diff = pd.Timedelta(seconds=(time.time() - start_time))
        logger.success("Model run finished. Runtime: " + str(time_diff))

    def run_single(self):
        """
        Process the next time step, i.e., increment the date counter and call the methods for
        preparing the meteorological fields and the interface to the submodules.
        """
        if self.date_idx is None:
            self.date_idx = 0
        elif self.date_idx == len(self.dates) - 1:
            raise errors.OpenAmundsenRuntimeError("Model run already finished")
        else:
            self.date_idx += 1

        self.date = self.dates[self.date_idx]

        logger.info(f"Processing time step {self.date:%Y-%m-%d %H:%M}")

        if self.config.reset_state_variables:
            self.state.reset()

        if self._require_interpolation:
            oameteo.interpolate_station_data(self)

        if self.config.input_data.meteo.format == "callback":
            self._meteo_callback(self)

        self._process_meteo_data()
        self._model_interface()
        self.point_output.update()
        self.gridded_output.update()

        if self.config.liveview.enabled:
            logger.debug("Updating live view window")
            self.liveview.update(self.date)

    def global_mask(self, mask, global_mask=None, global_idxs=None):
        if global_idxs is None:
            if global_mask is None:
                global_idxs = self.grid.roi_idxs_flat
            else:
                global_idxs = np.flatnonzero(global_mask)

        if mask.shape[-1] != len(global_idxs):
            raise ValueError("Local mask does not match global mask size")

        if mask.ndim == 1:
            global_mask = np.zeros(self.grid.shape, dtype=bool)
            idxs = global_idxs[mask]
            global_mask.flat[idxs] = True
        elif mask.ndim == 2:
            dim3 = mask.shape[0]
            global_mask = np.zeros((dim3, *self.grid.shape), dtype=bool)

            for i in range(dim3):
                idxs = global_idxs[mask[i, :]]
                global_mask[i, :, :].flat[idxs] = True

        return global_mask

    def roi_mask_to_global(self, mask):
        return self.global_mask(mask)

    def _prepare_time_steps(self):
        """
        Prepare the time steps for the model run according to the run
        configuration (start date, end date, and time step).
        """
        dates = pd.date_range(
            start=self.config["start_date"],
            end=self.config["end_date"],
            freq=self.config["timestep"],
        )
        self.dates = dates

        # Store timestep in seconds in the `timestep` attribute
        self.timestep = util.offset_to_timedelta(self.config["timestep"]).total_seconds()

        if self.config.simulation_timezone is None:
            hour_shift = 0
        else:
            hour_shift = self.config.timezone - self.config.simulation_timezone
        self._simulation_timezone_shift = pd.Timedelta(hours=hour_shift)

    def _model_interface(self):
        """
        Interface for calling the different submodules. This method is called
        in every time step after the meteorological fields have been prepared.
        """
        self._irradiance()

        if self._require_land_cover:
            self.land_cover.lai()

        if self._require_evapotranspiration or self._require_canopy:
            modules.canopy.above_canopy_meteorology(self)

        if self._require_canopy:
            self.canopy.meteorology()
            self.canopy.snow()

        self.snow.compaction()
        self.snow.accumulation()
        self.snow.albedo_aging()

        if self._require_snow_management:
            self.snow_management.produce()
            self.snow_management.groom()

        self.snow.update_properties()

        if self._require_soil:
            modules.soil.soil_properties(self)

        surface.surface_properties(self)

        if self._require_energy_balance:
            surface.energy_balance(self)
        elif self._require_temperature_index:
            self.snow.temperature_index()

        self.snow.heat_conduction()
        self.snow.melt()
        self.snow.sublimation()
        self.snow.runoff()
        self.snow.update_properties()
        self.snow.update_layers()

        if self._require_soil:
            modules.soil.soil_heat_flux(self)
            modules.soil.soil_temperature(self)

        if self._require_evapotranspiration:
            self.evapotranspiration.evapotranspiration()

        if self._require_glaciers:
            self.glaciers.update()

    def configure_logger(self):
        """
        Configure the logger.
        """
        # Remove all handlers and re-add default handler, filtering out openAMUNDSEN messages
        logger.remove()
        logger.add(sys.stderr, filter=lambda record: not record["name"].startswith("openamundsen."))

        # Add handler for openAMUNDSEN messages
        log_format = (
            "<green>{time:YYYY-MM-DD HH:mm:ss}</green> | <level>{level: <8}</level> | "
            + "<level>{message}</level>"
        )
        logger.add(
            sys.stderr,
            format=log_format,
            filter="openamundsen",
            level=self.config.log_level,
        )

    def _initialize_grid(self):
        """
        Initialize the grid parameters (number of rows and columns, transformation
        parameters) for the OpenAmundsen instance by reading the DEM file associated to the
        model run.
        """
        logger.info("Initializing model grid")

        dem_file = util.raster_filename("dem", self.config)
        meta = fileio.read_raster_metadata(dem_file, crs=self.config.crs)
        logger.info(f"Grid has dimensions {meta['rows']}x{meta['cols']}")

        grid = util.ModelGrid(meta)
        grid.prepare_coordinates()
        self.grid = grid

    def _initialize_state_variable_management(self):
        """
        Initialize the state variable management and add default state variables.
        """
        self.state = statevars.StateVariableManager(self.grid.rows, self.grid.cols)
        statevars.add_default_state_variables(self)

    def _create_state_variables(self):
        """
        Create the state variables (i.e., empty arrays with the shape of the
        model grid) for the model run.
        """
        self.state.initialize()

    def _initialize_state_variables(self):
        """
        Fill the state variables arrays with initial values.
        """
        self.snow.initialize()
        modules.soil.initialize(self)
        self.state.surface.temp[self.grid.roi] = self.state.soil.temp[0, self.grid.roi]

        if self._require_land_cover:
            self.land_cover.initialize()

        if self._require_canopy:
            self.canopy.initialize()

        if self._require_evapotranspiration:
            self.evapotranspiration.initialize()

        if self._require_snow_management:
            self.snow_management.initialize()

        if self._require_glaciers:
            self.glaciers.initialize()

    def _initialize_point_outputs(self):
        self.point_output = fileio.PointOutputManager(self)

    def _initialize_gridded_outputs(self):
        self.gridded_output = fileio.GriddedOutputManager(self)

    def _calculate_terrain_parameters(self):
        logger.info("Calculating terrain parameters")
        slope, aspect = terrain.slope_aspect(self.state.base.dem, self.grid.resolution)
        normal_vec = terrain.normal_vector(self.state.base.dem, self.grid.resolution)
        self.state.base.slope[:] = slope
        self.state.base.aspect[:] = aspect
        self.state.base.normal_vec[:] = normal_vec

        if self.config.meteo.interpolation.wind.method == "liston":
            # Calculate topographic curvature and normalize to a [-0.5, 0.5] range following Liston
            # et al. (2007)
            curv = terrain.curvature(
                self.state.base.dem,
                self.grid.resolution,
                "liston",
                L=self.config.meteo.interpolation.wind.curvature_length_scale,
            )
            self.state.base.scaled_curvature[:] = util.normalize_array(curv, -0.5, 0.5)

    def _read_input_data(self):
        """
        Read the input raster files required for the model run including the
        DEM, ROI (if available), and other files depending on the activated
        submodules.
        """
        dem_file = util.raster_filename("dem", self.config)
        roi_file = util.raster_filename("roi", self.config)
        svf_file = util.raster_filename("svf", self.config)

        self._read_extended_grids()

        # If an extendeded DEM+SVF are available, set the model DEM and SVF using them, otherwise
        # (try to) read them from file (or calculate SVF)
        if self.grid.extended_grid.available:
            self.state.base.dem[:] = self.grid.extended_grid.dem[
                self.grid.extended_grid.row_slice,
                self.grid.extended_grid.col_slice,
            ]
            self.state.base.svf[:] = self.grid.extended_grid.svf[
                self.grid.extended_grid.row_slice,
                self.grid.extended_grid.col_slice,
            ]
        else:
            if dem_file.exists():
                logger.info(f"Reading DEM ({dem_file})")
                self.state.base.dem[:] = fileio.read_raster_file(
                    dem_file,
                    check_meta=self.grid,
                    fill_value=np.nan,
                    dtype=float,
                )
            else:
                raise FileNotFoundError(f"DEM file not found: {dem_file}")

            if svf_file.exists():
                logger.info(f"Reading sky view factor ({svf_file})")
                self.state.base.svf[:] = fileio.read_raster_file(
                    svf_file,
                    check_meta=self.grid,
                    fill_value=np.nan,
                    dtype=float,
                )
            else:
                logger.info("Calculating sky view factor")
                svf = terrain.sky_view_factor(
                    self.state.base.dem,
                    self.grid.resolution,
                    num_sweeps=self.config.meteo.radiation.num_shadow_sweeps,
                )
                self.state.base.svf[:] = svf
                logger.debug(f"Writing sky view factor file ({svf_file})")
                fileio.write_raster_file(svf_file, svf, self.grid.transform, decimal_precision=3)

        if roi_file.exists():
            logger.info(f"Reading ROI ({roi_file})")
            self.grid.roi[:] = fileio.read_raster_file(
                roi_file,
                check_meta=self.grid,
                fill_value=False,
                dtype=bool,
            )
        else:
            logger.debug("No ROI file available, setting ROI to entire grid area")
            self.grid.roi[:] = True

        dem_nan_pos = np.isnan(self.state.base.dem) & self.grid.roi
        if np.any(dem_nan_pos):
            logger.debug(f"Excluding {dem_nan_pos.sum()} pixels where DEM is NaN from the ROI")
            self.grid.roi[dem_nan_pos] = False

        # Read snow redistribution factor files
        for precip_corr in self.config.meteo.precipitation_correction:
            if precip_corr["method"] == "srf":
                if "file" in precip_corr:
                    srf_file = precip_corr["file"]
                else:
                    srf_file = util.raster_filename("srf", self.config)

                logger.info(f"Reading snow redistribution factor ({srf_file})")
                self.state.base.srf[:] = fileio.read_raster_file(
                    srf_file,
                    check_meta=self.grid,
                    fill_value=np.nan,
                    dtype=float,
                )
                break

        # Read land cover file
        if self._require_land_cover:
            land_cover_file = util.raster_filename("lc", self.config)

            if land_cover_file.exists():
                logger.info(f"Reading land cover ({land_cover_file})")
                self.state.land_cover.land_cover[:] = fileio.read_raster_file(
                    land_cover_file,
                    check_meta=self.grid,
                    fill_value=0,
                    dtype=int,
                )
            else:
                raise FileNotFoundError(f"Land cover file not found: {land_cover_file}")

        # Read soil texture file
        if self._require_soil_texture:
            soil_texture_file = util.raster_filename("soil", self.config)

            if soil_texture_file.exists():
                logger.info(f"Reading soil texture ({soil_texture_file})")
                self.state.evapotranspiration.soil_texture[:] = fileio.read_raster_file(
                    soil_texture_file,
                    check_meta=self.grid,
                    fill_value=0,
                    dtype=int,
                )
            else:
                raise FileNotFoundError(f"Soil texture file not found: {soil_texture_file}")

        # Read glaciers file
        if self._require_glaciers:
            glaciers_file = util.raster_filename("glaciers", self.config)

            if glaciers_file.exists():
                logger.info(f"Reading glaciers ({glaciers_file})")
                self.state.glaciers.glacier[:] = fileio.read_raster_file(
                    glaciers_file,
                    check_meta=self.grid,
                )
            else:
                raise FileNotFoundError(f"Glaciers file not found: {glaciers_file}")

        self.grid.prepare_roi_coordinates()

    def _read_extended_grids(self):
        """
        Try to read the extended DEM if available, and read/calculate the extended SVF.
        """
        extended_dem_file = util.raster_filename("extended-dem", self.config)
        extended_svf_file = util.raster_filename("extended-svf", self.config)

        if not extended_dem_file.exists():
            return False

        logger.info(f"Reading extended DEM ({extended_dem_file})")
        ext_meta = fileio.read_raster_metadata(extended_dem_file, crs=self.config.crs)
        ext_dem = fileio.read_raster_file(
            extended_dem_file,
            fill_value=np.nan,
            dtype=float,
        )
        grid_transform = self.grid.transform
        ext_transform = ext_meta["transform"]
        grid_ul_xy = rasterio.transform.xy(grid_transform, 0, 0, offset="ul")
        grid_lr_xy = rasterio.transform.xy(
            grid_transform,
            self.grid.rows - 1,
            self.grid.cols - 1,
            offset="lr",
        )
        ext_ul_xy = rasterio.transform.xy(ext_transform, 0, 0, offset="ul")
        ext_lr_xy = rasterio.transform.xy(
            ext_transform,
            ext_meta["rows"] - 1,
            ext_meta["cols"] - 1,
            offset="lr",
        )
        ext_offset_ul = rasterio.transform.rowcol(ext_transform, *grid_ul_xy, op=float)
        if not (float.is_integer(ext_offset_ul[0]) and float.is_integer(ext_offset_ul[1])):
            raise errors.RasterFileError("Extended DEM is not aligned correctly")
        if not (
            grid_ul_xy[0] >= ext_ul_xy[0]
            and grid_ul_xy[1] <= ext_ul_xy[1]
            and grid_lr_xy[0] <= ext_lr_xy[0]
            and grid_lr_xy[1] >= ext_lr_xy[1]
        ):
            raise errors.RasterFileError("Extended DEM does not fully cover the model grid")

        if extended_svf_file.exists():
            ext_svf = fileio.read_raster_file(
                extended_svf_file,
                fill_value=np.nan,
                dtype=float,
            )
            if ext_svf.shape != ext_dem.shape:
                raise errors.RasterFileError("Extended DEM and SVF have differing dimensions")
        else:
            logger.info("Calculating extended sky view factor")
            ext_svf = terrain.sky_view_factor(
                ext_dem,
                self.grid.resolution,
                num_sweeps=self.config.meteo.radiation.num_shadow_sweeps,
            )
            logger.debug(f"Writing extended sky view factor file ({extended_svf_file})")
            fileio.write_raster_file(extended_svf_file, ext_svf, ext_transform, decimal_precision=3)

        row_offset = int(ext_offset_ul[0])
        col_offset = int(ext_offset_ul[1])
        self.grid.extended_grid.available = True
        self.grid.extended_grid.rows = ext_dem.shape[0]
        self.grid.extended_grid.cols = ext_dem.shape[1]
        self.grid.extended_grid.row_offset = row_offset
        self.grid.extended_grid.col_offset = col_offset
        self.grid.extended_grid.row_slice = slice(row_offset, row_offset + self.grid.rows)
        self.grid.extended_grid.col_slice = slice(col_offset, col_offset + self.grid.cols)
        self.grid.extended_grid.dem = ext_dem
        self.grid.extended_grid.svf = ext_svf
        self.grid.extended_grid.normal_vec = terrain.normal_vector(ext_dem, self.grid.resolution)

        return True

    def _read_meteo_data(self):
        """
        Read the meteorological data files required for the model run and store
        them in the `meteo` variable.
        """
        bounds = self.config.input_data.meteo.bounds
        if bounds == "grid":
            x_min = self.grid.x_min
            y_min = self.grid.y_min
            x_max = self.grid.x_max
            y_max = self.grid.y_max
        elif bounds == "global":
            x_min = -np.inf
            y_min = -np.inf
            x_max = np.inf
            y_max = np.inf
        elif isinstance(bounds, list):
            x_min, y_min, x_max, y_max = bounds

        if self.config.input_data.meteo.format == "csv":
            if self.config.input_data.meteo.crs is None:
                meteo_crs = self.config.crs
            else:
                meteo_crs = self.config.input_data.meteo.crs
        elif self.config.input_data.meteo.format == "netcdf":
            meteo_crs = None  # no CRS required for NetCDF input

        ds = fileio.read_meteo_data(
            self.config.input_data.meteo.format,
            self.config.input_data.meteo.dir,
            self.config.start_date,
            self.config.end_date,
            meteo_crs=meteo_crs,
            grid_crs=self.config.crs,
            bounds=(x_min, y_min, x_max, y_max),
            exclude=self.config.input_data.meteo.exclude,
            include=self.config.input_data.meteo.include,
            filters=self.config.input_data.meteo.filters,
            freq=self.config["timestep"],
            aggregate=self.config.input_data.meteo.aggregate_when_downsampling,
        )

        return ds

    def _process_meteo_data(self):
        """
        Calculate derived meteorological variables from the interpolated fields.
        """
        logger.debug("Calculating derived meteorological variables")

        m = self.state.meteo
        roi = self.grid.roi

        m.atmos_press[roi] = oameteo.atmospheric_pressure(self.state.base.dem[roi])
        m.sat_vap_press[roi] = oameteo.saturation_vapor_pressure(m.temp[roi])
        m.vap_press[roi] = oameteo.vapor_pressure(m.temp[roi], m.rel_hum[roi])
        m.spec_hum[roi] = oameteo.specific_humidity(m.atmos_press[roi], m.vap_press[roi])
        m.spec_heat_cap_moist_air[roi] = oameteo.specific_heat_capacity_moist_air(m.spec_hum[roi])
        m.lat_heat_vap[roi] = oameteo.latent_heat_of_vaporization(m.temp[roi])
        m.psych_const[roi] = oameteo.psychrometric_constant(
            m.atmos_press[roi],
            m.spec_heat_cap_moist_air[roi],
            m.lat_heat_vap[roi],
        )
        m.wet_bulb_temp[roi] = oameteo.wet_bulb_temperature(
            m.temp[roi],
            m.rel_hum[roi],
            m.vap_press[roi],
            m.psych_const[roi],
        )
        m.dew_point_temp[roi] = oameteo.dew_point_temperature(m.temp[roi], m.rel_hum[roi])
        m.precipitable_water[roi] = oameteo.precipitable_water(
            m.temp[roi],
            m.vap_press[roi],
        )
        m.dry_air_density[roi] = m.atmos_press[roi] / (constants.GAS_CONSTANT_DRY_AIR * m.temp[roi])

        # Calculate precipitation phase
        precip_phase_method = self.config.meteo.precipitation_phase.method
        if precip_phase_method == "temp":
            pp_temp = m.temp
        elif precip_phase_method == "wet_bulb_temp":
            pp_temp = m.wet_bulb_temp

        snowfall_frac = oameteo.precipitation_phase(
            pp_temp[roi],
            threshold_temp=self.config.meteo.precipitation_phase.threshold_temp,
            temp_range=self.config.meteo.precipitation_phase.temp_range,
        )
        m.snowfall[roi] = snowfall_frac * m.precip[roi]
        m.rainfall[roi] = (1 - snowfall_frac) * m.precip[roi]

        # Redistribute snow
        if "srf" in self.state.base:
            m.snowfall[roi] *= self.state.base.srf[roi]
            m.precip[roi] = m.snowfall[roi] + m.rainfall[roi]

    def _irradiance(self):
        self.sun_params = modules.radiation.sun_parameters(
            self.date + self._simulation_timezone_shift,
            self.grid.center_lon,
            self.grid.center_lat,
            self.config.timezone,
        )

        modules.radiation.clear_sky_shortwave_irradiance(self)

        if self._require_interpolation:
            modules.radiation.shortwave_irradiance(self)
        else:
            m = self.state.meteo
            roi = self.grid.roi
            m.sw_in[roi] = m.sw_in_clearsky[roi] * m.cloud_factor[roi]

        modules.radiation.longwave_irradiance(self)

    @property
    def timestep_props(self):
        """
        Get the properties of the current timestep.

        The return value is a dataclass with the following fields:

            - `first_of_run`: Whether this is the first timestep of the model
              run.
            - `strict_first_of_year`: Whether this is the first possible
              timestep of the current year.
            - `strict_first_of_month`: Whether this is the first possible
              timestep of the current month.
            - `strict_first_of_day`: Whether this is the first possible timestep
              of the current day.
            - `first_of_year`: Whether this is the first actually processed
              timestep of the current year.
            - `first_of_month`: Whether this is the first actually processed
              timestep of the current month.
            - `first_of_day`: Whether this is the first actually processed
              timestep of the current day.
            - `last_of_run`: Whether this is the last timestep of the model run.
            - `strict_last_of_year`: Whether this is the last possible timestep
              of the current year.
            - `strict_last_of_month`: Whether this is the last possible timestep
              of the current month.
            - `strict_last_of_day`: Whether this is the last possible timestep
              of the current day.
            - `last_of_year`: Whether this is the last actually processed
              timestep of the current year.
            - `last_of_month`: Whether this is the last actually processed
              timestep of the current month.
            - `last_of_day`: Whether this is the last actually processed
              timestep of the current day.

        The difference between the "strict" and the "non-strict" fields is that
        the latter take the model run start/end date into account. For example,
        if the start date is 2020-12-31 23:00, at the first timestep
        `first_of_year`, `first_of_month` and `first_of_day` will all be True
        (since these are the first _processed_ timesteps of the current year,
        month and day), whereas `strict_first_of_year`, `strict_first_of_month`
        and `strict_first_of_day` will only be True for the first time on
        2021-01-01 00:00.
        """
        pot_prev_date = self.date - pd.Timedelta(seconds=self.timestep)
        pot_next_date = self.date + pd.Timedelta(seconds=self.timestep)

        first_of_run = self.date_idx == 0
        strict_first_of_year = pot_prev_date.year != self.date.year
        strict_first_of_month = pot_prev_date.month != self.date.month
        strict_first_of_day = pot_prev_date.day != self.date.day
        first_of_year = strict_first_of_year or first_of_run
        first_of_month = strict_first_of_month or first_of_run
        first_of_day = strict_first_of_day or first_of_run

        last_of_run = self.date_idx == len(self.dates) - 1
        strict_last_of_year = pot_next_date.year != self.date.year
        strict_last_of_month = pot_next_date.month != self.date.month
        strict_last_of_day = pot_next_date.day != self.date.day
        last_of_year = strict_last_of_year or last_of_run
        last_of_month = strict_last_of_month or last_of_run
        last_of_day = strict_last_of_day or last_of_run

        return util.TimestepProperties(
            first_of_run=first_of_run,
            strict_first_of_year=strict_first_of_year,
            strict_first_of_month=strict_first_of_month,
            strict_first_of_day=strict_first_of_day,
            first_of_year=first_of_year,
            first_of_month=first_of_month,
            first_of_day=first_of_day,
            last_of_run=last_of_run,
            strict_last_of_year=strict_last_of_year,
            strict_last_of_month=strict_last_of_month,
            strict_last_of_day=strict_last_of_day,
            last_of_year=last_of_year,
            last_of_month=last_of_month,
            last_of_day=last_of_day,
        )


def Model(*args, **kwargs):
    import warnings

    warnings.warn(
        "Using oa.Model is deprecated, please use oa.OpenAmundsen instead",
        DeprecationWarning,
        stacklevel=2,
    )
    return OpenAmundsen(*args, **kwargs)



import warnings

import numpy as np
import pytest
import xarray as xr


def compare_datasets(
    name,
    ds_dev,
    point=None,
    variables=None,
    data_dir=pytest.COMPARISON_DATA_DIR,
    prepare_comparison_data=pytest.PREPARE_COMPARISON_DATA,
    reports_dir=pytest.REPORTS_DIR,
):
    plot = reports_dir is not None
    nc_file = data_dir / f"{name}.nc"

    if prepare_comparison_data:
        ds_dev.to_netcdf(nc_file)
        return

    ds_base = xr.load_dataset(nc_file)

    if point is not None:
        ds_base = ds_base.sel(point=point)
        ds_dev = ds_dev.sel(point=point)

    # Check if we are comparing time series or gridded results
    is_time_series = "point" in ds_base.coords

    if variables is None:
        base_vars = [v for v in ds_base.variables if v not in ds_base.coords]
        dev_vars = [v for v in ds_dev.variables if v not in ds_dev.coords]
        variables = base_vars
        variables.extend([v for v in dev_vars if v not in variables])

    compare_vars = []
    for v in variables:
        if not (v in ds_base.variables and v in ds_dev.variables):
            warnings.warn(f"Variable {v} not in both datasets, cannot compare", stacklevel=2)
        elif ds_base[v].shape != ds_dev[v].shape:
            warnings.warn(f"Non-matching shapes for variable {v}", stacklevel=2)
        else:
            compare_vars.append(v)

    if plot:
        plot_vars = []

        for v in compare_vars:
            if len(ds_base[v].dims) > 3:
                warnings.warn(
                    f"Variable {v} is 3-dimensional, plotting not supported",
                    stacklevel=2,
                )
            else:
                plot_vars.append(v)

        if is_time_series:
            changed_vars = []
        else:
            changed_vars = {}

    for v in compare_vars:
        data_base = ds_base[v].values
        data_dev = ds_dev[v].values

        if is_time_series:
            if not np.allclose(data_base, data_dev, equal_nan=True):
                max_abs_diff, max_rel_diff = _max_diff(data_base, data_dev)
                warnings.warn(
                    f"Non-matching values for variable {v}. "
                    f"Max abs/rel diff: {max_abs_diff:g} / {max_rel_diff:g}",
                    stacklevel=2,
                )

                if plot:
                    changed_vars.append(v)
        else:
            time_dim = ds_base[v].dims[0]
            time_vals = ds_base[v].coords[time_dim].to_index()

            for date_num in range(data_base.shape[0]):
                data_base_cur = data_base[date_num, :, :]
                data_dev_cur = data_dev[date_num, :, :]
                date = time_vals[date_num]

                if not np.allclose(data_base_cur, data_dev_cur, equal_nan=True):
                    max_abs_diff, max_rel_diff = _max_diff(data_base_cur, data_dev_cur)
                    warnings.warn(
                        f"Non-matching values for variable {v} ({date}). "
                        f"Max abs/rel diff: {max_abs_diff:g} / {max_rel_diff:g}",
                        stacklevel=2,
                    )

                    if plot:
                        if v not in changed_vars:
                            changed_vars[v] = []

                        changed_vars[v].append(date_num)

    if plot:
        from .plot import fig_to_html, plot_gridded_comparison, plot_point_comparison
        # (import here because plotly is only required for report generation)

        if is_time_series:
            fig = plot_point_comparison(
                ds_base,
                ds_dev,
                plot_vars,
                changed_vars,
            )
        else:
            fig = plot_gridded_comparison(
                ds_base,
                ds_dev,
                plot_vars,
                changed_vars,
            )

        fig_to_html(fig, reports_dir / f"{name}.html")


def _max_diff(x, y):
    pos = np.isfinite(x) & np.isfinite(y)
    x = x[pos]
    y = y[pos]
    nonzero = x != 0
    abs_diff = np.abs(x - y)
    max_abs_diff = np.max(abs_diff)
    max_rel_diff = np.max(abs_diff[nonzero] / np.abs(x[nonzero]))
    return max_abs_diff, max_rel_diff



import warnings

import numpy as np
import pytest
from numpy.testing import assert_allclose, assert_equal

import openamundsen as oa

from .compare import compare_datasets
from .conftest import base_config

single_point_results_all = [
    "single_point_results_multilayer",
    "single_point_results_cryolayers",
]


def base_config_snow():
    config = base_config()
    config.start_date = "2019-10-01"
    config.end_date = "2020-05-31"
    config.output_data.timeseries.variables = [
        {"var": "snow.num_layers", "name": "num_snow_layers"},
        {"var": "snow.albedo", "name": "snow_albedo"},
    ]
    return config


@pytest.fixture(scope="session")
def multilayer_run():
    config = base_config_snow()
    config.snow.model = "multilayer"
    model = oa.OpenAmundsen(config)
    model.initialize()
    model.run()
    return model


@pytest.fixture(scope="session")
def cryolayer_run():
    config = base_config_snow()
    config.snow.model = "cryolayers"
    config.output_data.timeseries.variables.append({"var": "snow.cold_content"})
    config.output_data.timeseries.variables.append({"var": "snow.layer_albedo"})
    model = oa.OpenAmundsen(config)
    model.initialize()
    model.run()
    return model


@pytest.fixture(scope="module")
def point_results_cryolayers():
    config = base_config()
    config.snow.model = "cryolayers"
    config.output_data.timeseries.variables.append({"var": "snow.cold_content"})
    config.output_data.timeseries.variables.append({"var": "snow.layer_albedo"})
    model = oa.OpenAmundsen(config)
    model.initialize()
    model.run()
    return model.point_output.data


@pytest.fixture(scope="function")
def single_point_results_multilayer(multilayer_run):
    return multilayer_run.point_output.data.sel(point="proviantdepot")


@pytest.fixture(scope="function")
def single_point_results_cryolayers(cryolayer_run):
    return cryolayer_run.point_output.data.sel(point="proviantdepot")


def test_default_multilayer():
    config = base_config_snow()
    assert config.snow.model == "multilayer"


@pytest.mark.slow
@pytest.mark.parametrize("fixture_name", single_point_results_all)
def test_swe(fixture_name, request):
    ds = request.getfixturevalue(fixture_name)
    swe = ds.swe.values
    ice_content = ds.ice_content.values
    liquid_water_content = ds.liquid_water_content.values
    assert np.all(np.isfinite(swe))
    assert swe.min() == 0.0
    assert_allclose(
        swe,
        (ice_content + liquid_water_content).sum(axis=1),
        atol=0.1,
    )
    assert np.all(ice_content >= 0.0)
    assert np.all(liquid_water_content >= 0.0)


@pytest.mark.slow
@pytest.mark.parametrize("fixture_name", single_point_results_all)
def test_density(fixture_name, request):
    ds = request.getfixturevalue(fixture_name)
    swe3d = (ds.ice_content + ds.liquid_water_content).values
    density = ds.snow_density.values
    assert np.all(np.isnan(density[swe3d == 0.0]))
    assert np.all(np.isfinite(density[swe3d > 0.0]))
    assert not np.any(density < 0.0)
    assert not np.any(density > 1000.0)


@pytest.mark.slow
@pytest.mark.parametrize("fixture_name", single_point_results_all)
def test_depth(fixture_name, request):
    ds = request.getfixturevalue(fixture_name)
    depth = ds.snow_depth.values
    swe = ds.swe.values
    assert np.all(depth[swe == 0.0] == 0.0)
    assert np.all(depth[swe > 0.0] > 0.0)
    assert not np.any(depth < 0.0)

    swe3d = (ds.ice_content + ds.liquid_water_content).values
    thickness = ds.snow_thickness.values
    assert_allclose(depth, thickness.sum(axis=1), atol=1e-3)

    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            "invalid value encountered in divide",
            RuntimeWarning,
        )
        density_calc = swe3d / thickness

    density = ds.snow_density.values
    assert_allclose(density_calc, density, atol=1.0)


@pytest.mark.slow
@pytest.mark.parametrize("fixture_name", single_point_results_all)
def test_num_layers(fixture_name, request):
    ds = request.getfixturevalue(fixture_name)
    num_layers = ds.num_snow_layers.values
    thickness = ds.snow_thickness.values
    assert_equal(num_layers, (thickness > 0).sum(axis=1))


@pytest.mark.slow
@pytest.mark.parametrize("fixture_name", single_point_results_all)
def test_albedo(fixture_name, request):
    ds = request.getfixturevalue(fixture_name)
    config = base_config()
    albedo = ds.snow_albedo.values
    swe = ds.swe.values
    min_albedo = config.snow.albedo.min
    max_albedo = config.snow.albedo.max
    pos_snow = swe > 0.0
    assert np.all(albedo[pos_snow] >= np.float32(min_albedo))
    assert np.all(albedo[pos_snow] <= np.float32(max_albedo))
    assert np.all(np.isnan(albedo[~pos_snow]))


@pytest.mark.slow
@pytest.mark.parametrize("fixture_name", single_point_results_all)
def test_melt(fixture_name, request):
    ds = request.getfixturevalue(fixture_name)
    assert np.all(ds.melt >= 0)


@pytest.mark.slow
@pytest.mark.parametrize("fixture_name", single_point_results_all)
def test_runoff(fixture_name, request):
    ds = request.getfixturevalue(fixture_name)
    assert np.all(ds.runoff >= 0)


@pytest.mark.slow
@pytest.mark.parametrize("fixture_name", single_point_results_all)
def test_refreezing(fixture_name, request):
    ds = request.getfixturevalue(fixture_name)
    assert np.all(ds.refreezing >= 0)


@pytest.mark.slow
def test_cryolayers_layer_albedo(single_point_results_cryolayers):
    config = base_config()
    ds = single_point_results_cryolayers
    layer_albedo = ds.layer_albedo.values
    swe3d = (ds.ice_content + ds.liquid_water_content).values
    min_albedo = config.snow.albedo.min
    max_albedo = config.snow.albedo.max
    pos_snow = swe3d > 0.0
    assert np.all(layer_albedo[pos_snow] >= np.float32(min_albedo))
    assert np.all(layer_albedo[pos_snow] <= np.float32(max_albedo))
    assert np.all(np.isnan(layer_albedo[~pos_snow]))


@pytest.mark.slow
def test_cryolayers_cold_content(single_point_results_cryolayers):
    config = base_config()
    ds = single_point_results_cryolayers
    cc = ds.cold_content.values
    ic = ds.ice_content.values
    lwc = ds.liquid_water_content.values
    melt = ds.melt.values
    sublimation = ds.sublimation.values
    swe3d = (ds.ice_content + ds.liquid_water_content).values
    cold_holding_capacity = config.snow.cryolayers.cold_holding_capacity
    pos_snow = swe3d > 0
    assert np.all(cc >= 0.0)
    assert cc.max() > 0.0
    assert np.all(cc[~pos_snow] == 0.0)

    # Check if the CC is <= swe * cold_colding_capacity (add melt and sublimation because the
    # maximum CC is calculated before reducing SWE by melt and sublimation)
    assert np.all(
        (((ic + lwc).sum(axis=1) + melt + sublimation) * cold_holding_capacity - cc.sum(axis=1))
        >= -0.1
    )


@pytest.mark.slow
@pytest.mark.parametrize("method", ["temperature_index", "enhanced_temperature_index"])
def test_temperature_index(method):
    config = base_config_snow()
    config.snow.model = "cryolayers"
    config.timestep = "D"
    config.output_data.timeseries.variables = [{"var": "snow.cold_content"}]
    config.snow.melt.method = method

    if method == "temperature_index":
        config.snow.melt.degree_day_factor = 5.0
    elif method == "enhanced_temperature_index":
        config.snow.melt.degree_day_factor = 3.0
        config.snow.melt.albedo_factor = 0.1

    model = oa.OpenAmundsen(config)
    model.initialize()
    model.run()

    ds = model.point_output.data.sel(point="proviantdepot")
    assert np.all(ds.melt >= 0)
    assert np.any(ds.melt > 0)
    assert np.all(ds.melt[ds.temp <= config.snow.melt.degree_day_factor] == 0.0)
    melt_potential = (
        ds.ice_content.sum("snow_layer").shift(time=1)
        + ds.cold_content.sum("snow_layer").shift(time=1)
        + ds.snowfall
    )
    assert np.all(ds.melt <= melt_potential.fillna(np.inf) + 1e-6)


@pytest.mark.slow
@pytest.mark.comparison
def test_compare_multilayer(multilayer_run):
    compare_datasets(
        "snow_multilayer_point",
        multilayer_run.point_output.data,
        point="proviantdepot",
    )


@pytest.mark.slow
@pytest.mark.comparison
def test_compare_cryolayers(cryolayer_run):
    compare_datasets(
        "snow_cryolayers_point",
        cryolayer_run.point_output.data,
        point="proviantdepot",
    )


@pytest.mark.slow
@pytest.mark.parametrize("fixture_name", single_point_results_all)
def test_mass_balance(fixture_name, request):
    ds = request.getfixturevalue(fixture_name)
    ds_prev = ds.shift(time=1)

    ic1 = ds.ice_content.sum("snow_layer")
    ic2 = (
        ds_prev.ice_content.sum("snow_layer")
        + ds.snowfall
        - ds.melt
        - ds.sublimation
        + ds.refreezing
    )
    assert_allclose(ic1, ic2, atol=1e-4)

    lwc1 = ds.liquid_water_content.sum("snow_layer")
    lwc2 = (
        ds_prev.liquid_water_content.sum("snow_layer")
        + ds.rainfall
        + ds.melt
        - ds.runoff
        - ds.refreezing
    )
    assert_allclose(lwc1, lwc2, atol=1e-4)



import multiprocessing

import openamundsen as oa

from .conftest import base_config


def run_model(model):
    model.configure_logger()
    model.run()


def test_multiprocessing():
    config = base_config()
    config.end_date = "2020-01-16"
    config.output_data.timeseries.format = "memory"
    config.output_data.grids.format = "memory"

    config.snow.model = "multilayer"
    model1 = oa.OpenAmundsen(config)
    model1.initialize()

    config.snow.model = "cryolayers"
    model2 = oa.OpenAmundsen(config)
    model2.initialize()

    spawn_context = multiprocessing.get_context("spawn")
    with spawn_context.Pool() as pool:
        pool.map(run_model, [model1, model2])



from dataclasses import dataclass

import numpy as np
from munch import Munch

from openamundsen import errors

_DTYPE_INIT_VALS = {
    float: np.nan,
    int: 0,
    bool: False,
}
_DTYPE_INIT_VALS[np.dtype("float64")] = _DTYPE_INIT_VALS[float]
_DTYPE_INIT_VALS[np.dtype("int64")] = _DTYPE_INIT_VALS[int]


class StateVariableManager:
    """
    Class for managing state variables of a OpenAmundsen instance.
    State variables are organized into categories (e.g., "base", "meteo",
    "snow").  For each category, a StateVariableContainer is created containing
    the variables and their respective metadata.

    Parameters
    ----------
    rows, cols : int
        Number of rows and columns of the model grid.
    """

    def __init__(self, rows, cols):
        self._rows = rows
        self._cols = cols
        self._categories = []

    @property
    def categories(self):
        """
        Return the currently assigned categories.
        """
        return tuple(self._categories)

    def add_category(self, category):
        """
        Add a category.

        Parameters
        ----------
        category : str
            Category name.
        """
        if category in self.categories:
            raise errors.CategoryError(f"Category {category} already exists")

        if not category.isidentifier():
            raise errors.CategoryError("Category name must be a valid Python identifier")

        self._categories.append(category)
        self._categories.sort()
        svc = StateVariableContainer(self)
        setattr(self, category, svc)

        return svc

    def parse(self, s):
        """
        Parse a string in the form "<category>[.<variable>]" and return the
        respective category and variable name (or None for the latter, it not
        specified).

        Returns
        -------
        (category, var_name): (str, str or None) tuple
        """
        if "." in s:
            category, var_name = s.split(".")
        else:
            category = s
            var_name = None

        return category, var_name

    def __getitem__(self, key):
        """
        Return either the StateVariableContainer for a category (if key is a
        category identifier) or a variable itself (if key is a string in the
        form "<category>.<variable>".

        Parameters
        ----------
        key : str

        Returns
        -------
        StateVariableContainer or ndarray
        """
        category, var_name = self.parse(key)

        if var_name is None:
            return getattr(self, category)
        else:
            return getattr(self, category)[var_name]

    def initialize(self):
        """
        Initialize the fields (i.e., create the arrays) for all variables of
        all categories.
        """
        for category in self.categories:
            svc = self[category]

            for var_name, var_def in svc._meta.items():
                if var_def.dim3 == 0:
                    arr = create_empty_array((self._rows, self._cols), var_def.dtype)
                else:
                    arr = create_empty_array((var_def.dim3, self._rows, self._cols), var_def.dtype)

                svc[var_name] = arr

    def reset(self):
        """
        Fill all state variables with their default "no data" value, except
        those for which the `retain` keyword is set.
        """
        for category in self.categories:
            svc = self[category]

            for var_name, var_def in svc._meta.items():
                if not var_def.retain:
                    arr = svc[var_name]
                    arr.fill(_DTYPE_INIT_VALS[arr.dtype])

    def meta(self, var):
        """
        Return metadata of a state variable.

        Parameters
        ----------
        var : str
            Variable name, e.g. "meteo.temp".

        Returns
        -------
        StateVariableDefinition
        """
        category, var_name = self.parse(var)
        return self[category]._meta[var_name]


class StateVariableContainer(Munch):
    """
    Container for storing state variables. This class inherits from `Munch` so
    that attributes are accessible both using dict notation (`svc['temp']`)
    as well as dot notation (`svc.temp`).

    Parameters
    ----------
    manager : StateVariableManager
        StateVariableManager instance to which this container is associated.
    """

    def __init__(self, manager):
        self._manager = manager
        self._meta = {}

    def add_variable(
        self,
        name,
        units=None,
        long_name=None,
        standard_name=None,
        dtype=float,
        dim3=0,
        retain=False,
    ):
        """
        Add a variable along with optional metadata.
        Note that the variable itself is not yet created here (this is done in
        StateVariableManager.iniitalize); only its name and attribute are
        stored.

        Parameters
        ----------
        name : str
            Name of the variable to be created. Must be a valid Python identifier.

        units : str, optional
            udunits-compatible definition of the units of the given variable.

        long_name : str, optional
            Long descriptive name of the variable.

        standard_name : str, optional
            Standard name of the variable in CF notation
            (http://cfconventions.org/standard-names.html).

        dtype : type, default float
            Data type for the variable values.

        dim3 : int, default 0
            Size of an optional third dimension of the field.

        retain : bool, default False
            Whether the variable contents should be retained or it is safe to
            reset the variable in every timestep.
        """
        definition = StateVariableDefinition(
            units=units,
            long_name=long_name,
            standard_name=standard_name,
            dtype=dtype,
            dim3=dim3,
            retain=retain,
        )

        self._meta[name] = definition


@dataclass(frozen=True)
class StateVariableDefinition:
    """
    Class for describing metadata of a state variable.

    Parameters
    ----------
    units : str, optional
        udunits-compatible definition of the units of the given variable.

    long_name : str, optional
        Long descriptive name of the variable.

    standard_name : str, optional
        Standard name of the variable in CF notation
        (http://cfconventions.org/standard-names.html).

    dtype : type, default float
        Data type for the variable values.

    dim3 : int, default 0
        Size of an optional third dimension of the field.

    retain : bool, default False
        Whether the variable contents should be retained or it is safe to reset
        the variable in every timestep.

    Examples
    --------
    >>> StateVariableDefinition(
            standard_name='precipitation_flux',
            long_name='Total precipitation flux',
            units='kg m-2 s-1')
    """

    units: str = None
    long_name: str = None
    standard_name: str = None
    dtype: type = float
    dim3: int = 0
    retain: bool = False


def add_default_state_variables(model):
    """
    Add all state variables to an OpenAmundsen instance which are required for any
    model run. Depending on which submodules are activated in the run
    configuration, further state variables might be added in other locations.
    """
    state = model.state

    # Base variables
    base = state.add_category("base")
    base.add_variable("dem", "m", "Surface altitude", "surface_altitude", retain=True)
    base.add_variable("slope", "degree", "Terrain slope", retain=True)
    base.add_variable("aspect", "degree", "Terrain aspect", retain=True)
    base.add_variable("normal_vec", long_name="Vector normal to the surface", dim3=3, retain=True)
    base.add_variable("svf", long_name="Sky-view factor", retain=True)

    # Meteorological variables
    meteo = state.add_category("meteo")
    meteo.add_variable("temp", "K", "Air temperature", "air_temperature")
    meteo.add_variable("precip", "kg m-2", "Precipitation amount", "precipitation_amount")
    meteo.add_variable("snowfall", "kg m-2", "Snowfall amount", "snowfall_amount")
    meteo.add_variable("rainfall", "kg m-2", "Rainfall amount", "rainfall_amount")
    meteo.add_variable("rel_hum", "%", "Relative humidity", "relative_humidity")
    meteo.add_variable("wind_speed", "m s-1", "Wind speed", "wind_speed")
    meteo.add_variable("wind_speed_gust", "m s-1", "Wind speed of gust", "wind_speed_of_gust")
    meteo.add_variable("wind_dir", "degree", "Wind direction", "wind_from_direction")
    meteo.add_variable(
        "sw_in",
        "W m-2",
        "Incoming shortwave radiation",
        "surface_downwelling_shortwave_flux_in_air",
    )
    meteo.add_variable(
        "sw_out",
        "W m-2",
        "Outgoing shortwave radiation",
        "surface_upwelling_shortwave_flux_in_air",
    )
    meteo.add_variable(
        "lw_in",
        "W m-2",
        "Incoming longwave radiation",
        "downwelling_longwave_flux_in_air",
    )
    meteo.add_variable(
        "lw_out",
        "W m-2",
        "Outgoing longwave radiation",
        "surface_upwelling_longwave_flux_in_air",
    )
    meteo.add_variable(
        "net_radiation",
        "W m-2",
        "Net radiation",
        "surface_net_downward_radiative_flux",
    )
    meteo.add_variable(
        "sw_in_clearsky",
        "W m-2",
        "Clear-sky incoming shortwave radiation",
        "surface_downwelling_shortwave_flux_in_air_assuming_clear_sky",
    )
    meteo.add_variable("dir_in_clearsky", "W m-2", "Clear-sky direct incoming shortwave radiation")
    meteo.add_variable(
        "diff_in_clearsky",
        "W m-2",
        "Clear-sky diffuse incoming shortwave radiation",
        "surface_diffuse_downwelling_shortwave_flux_in_air_assuming_clear_sky",
    )
    meteo.add_variable("cloud_factor", "1", "Cloud factor", retain=True)
    meteo.add_variable("cloud_fraction", "1", "Cloud fraction", "cloud_area_fraction")
    meteo.add_variable("wet_bulb_temp", "K", "Wet-bulb temperature", "wet_bulb_temperature")
    meteo.add_variable("dew_point_temp", "K", "Dew point temperature", "dew_point_temperature")
    meteo.add_variable("atmos_press", "Pa", "Atmospheric pressure", "air_pressure")
    meteo.add_variable("sat_vap_press", "Pa", "Saturation vapor pressure")
    meteo.add_variable("vap_press", "Pa", "Vapor pressure", "water_vapor_partial_pressure_in_air")
    meteo.add_variable("spec_hum", "kg kg-1", "Specific humidity", "specific_humidity")
    meteo.add_variable(
        "spec_heat_cap_moist_air",
        "J kg-1 K-1",
        "Specific heat capacity of moist air",
    )
    meteo.add_variable("psych_const", "Pa K-1", "Psychrometric constant")
    meteo.add_variable("lat_heat_vap", "J kg-1", "Latent heat of vaporization")
    meteo.add_variable("precipitable_water", "kg m-2", "Precipitable water")
    meteo.add_variable("dry_air_density", "kg m-3", "Dry air density")
    if model._require_evapotranspiration or model._require_canopy:
        meteo.add_variable("top_canopy_temp", "K", "Above-canopy air temperature")
        meteo.add_variable("top_canopy_rel_hum", "%", "Above-canopy relative humidity")
        meteo.add_variable("top_canopy_wind_speed", "m s-1", "Above-canopy wind speed")
        meteo.add_variable("top_canopy_sw_in", "W m-2", "Above-canopy incoming shortwave radiation")
        meteo.add_variable("top_canopy_lw_in", "W m-2", "Above-canopy incoming longwave radiation")

    # Surface variables
    surf = state.add_category("surface")
    surf.add_variable("temp", "K", "Surface temperature", "surface_temperature", retain=True)
    surf.add_variable("albedo", "1", "Surface albedo", "surface_albedo", retain=True)
    surf.add_variable("heat_flux", "W m-2", "Surface heat flux")
    surf.add_variable(
        "sens_heat_flux",
        "W m-2",
        "Sensible heat flux",
        "surface_downward_sensible_heat_flux",
    )
    surf.add_variable(
        "lat_heat_flux",
        "W m-2",
        "Latent heat flux",
        "surface_downward_latent_heat_flux",
    )
    surf.add_variable("moisture_flux", "kg m-2 s-1", "Moisture flux")
    surf.add_variable("advective_heat_flux", "W m-2", "Heat advected by precipitation")
    surf.add_variable("lat_heat", "J kg-1", "Latent heat")
    surf.add_variable(
        "sat_spec_hum",
        "kg kg-1",
        "Saturation specific humidity at surface temperature",
    )
    surf.add_variable("moisture_availability", "1", "Moisture availability factor")
    surf.add_variable(
        "roughness_length",
        "m",
        "Surface roughness length",
        "surface_roughness_length",
    )
    surf.add_variable("turbulent_exchange_coeff", "1", "Transfer coefficient for heat and moisture")
    surf.add_variable("conductance", "m s-1", "Surface conductance")
    if model.config.snow.model == "multilayer":
        surf.add_variable("layer_temp", "K", "Surface layer temperature", retain=True)
        surf.add_variable("thickness", "m", "Surface layer thickness", retain=True)
        surf.add_variable("therm_cond", "W m-1 K-1", "Surface thermal conductivity")
    elif model.config.snow.model == "cryolayers":
        surf.add_variable("layer_type", "1", "Surface layer type", dtype=int)

    # Snow variables (shared by all snow models, additional ones might be added from the individual
    # models)
    snow = state.add_category("snow")
    snow.add_variable("swe", "kg m-2", "Snow water equivalent", "surface_snow_amount", retain=True)
    snow.add_variable("depth", "m", "Snow depth", "surface_snow_thickness")
    snow.add_variable("melt", "kg m-2", "Snow melt", "surface_snow_melt_amount")
    snow.add_variable(
        "sublimation",
        "kg m-2",
        "Snow sublimation",
        "surface_snow_sublimation_amount",
    )
    snow.add_variable("refreezing", "kg m-2", "Liquid water refreezing")
    snow.add_variable("runoff", "kg m-2", "Snow runoff")
    snow.add_variable("albedo", "1", "Snow albedo", retain=True)
    snow.add_variable("area_fraction", "1", "Snow cover fraction", "surface_snow_area_fraction")

    # Soil variables
    soil = state.add_category("soil")
    num_soil_layers = len(model.config.soil.thickness)
    soil.add_variable("heat_flux", "W m-2", "Soil heat flux", "downward_heat_flux_in_soil")
    soil.add_variable(
        "vol_heat_cap_dry",
        "J K-1 m-3",
        "Volumetric heat capacity of dry soil",
        retain=True,
    )
    soil.add_variable("sat_water_pressure", "m", "Saturated soil water pressure", retain=True)
    soil.add_variable(
        "vol_moisture_content_sat",
        "m3 m-3",
        "Volumetric soil moisture content at saturation",
        retain=True,
    )
    soil.add_variable(
        "vol_moisture_content_crit",
        "m3 m-3",
        "Volumetric soil moisture content at critical point",
        retain=True,
    )
    soil.add_variable(
        "frac_frozen_moisture_content",
        "1",
        "Fractional frozen soil moisture content",
        dim3=num_soil_layers,
        retain=True,
    )
    soil.add_variable(
        "frac_unfrozen_moisture_content",
        "1",
        "Fractional unfrozen soil moisture content",
        dim3=num_soil_layers,
        retain=True,
    )
    soil.add_variable("clapp_hornberger", "1", "Clapp-Hornberger exponent", retain=True)
    soil.add_variable(
        "vol_moisture_content",
        "m3 m-3",
        "Volumetric soil moisture content",
        dim3=num_soil_layers,
        retain=True,
    )
    soil.add_variable("temp", "K", "Soil temperature", dim3=num_soil_layers, retain=True)
    soil.add_variable("thickness", "m", "Soil thickness", dim3=num_soil_layers, retain=True)
    soil.add_variable("heat_cap", "J K-1 m-2", "Areal heat capacity of soil", dim3=num_soil_layers)
    soil.add_variable(
        "therm_cond",
        "W m-1 K-1",
        "Thermal conductivity of soil",
        dim3=num_soil_layers,
    )
    soil.add_variable(
        "therm_cond_minerals",
        "W m-1 K-1",
        "Thermal conductivity of soil minerals",
        retain=True,
    )
    soil.add_variable(
        "therm_cond_dry",
        "W m-1 K-1",
        "Thermal conductivity of dry soil",
        retain=True,
    )


def create_empty_array(shape, dtype):
    """
    Create an empty array with a given shape and dtype initialized to "no
    data". The value of "no data" depends on the dtype and is e.g. NaN for
    float, 0 for int and False for float.

    Parameters
    ----------
    shape : int or sequence of ints
        Shape of the new array, e.g. (2, 3).

    dtype : type
        The desired data type for the array.

    Returns
    -------
    out : ndarray
    """
    return np.full(shape, _DTYPE_INIT_VALS[dtype], dtype=dtype)



from pathlib import Path

import numpy as np
import pytest
import rasterio
import rasterio.windows
import xarray as xr
from numpy.testing import assert_allclose, assert_equal

import openamundsen as oa
import openamundsen.errors as errors

from .conftest import base_config


def base_cloudiness_config():
    config = base_config()
    config.start_date = "2019-10-05"
    config.end_date = "2019-10-10"
    config.input_data.meteo.bounds = "global"
    return config


def test_clear_sky_fraction():
    nan_pos = slice("2019-10-06 00:00", "2019-10-07 00:00")

    config = base_cloudiness_config()
    config.meteo.interpolation.cloudiness.method = "clear_sky_fraction"
    config.meteo.interpolation.cloudiness.allow_fallback = False
    model = oa.OpenAmundsen(config)
    model.initialize()

    model.meteo.sw_in.loc["proviantdepot", nan_pos] = np.nan
    model.run()
    ds = model.point_output.data
    assert np.all(ds.sw_in.notnull())

    model = oa.OpenAmundsen(config)
    model.initialize()
    model.meteo.sw_in.loc[:, nan_pos] = np.nan
    model.run()
    ds = model.point_output.data
    # When allow_fallback=False, values are nan during daytime
    assert np.all(ds.sw_in.sel(time=slice("2019-10-06 12:00", "2019-10-06 15:00")).isnull())
    assert np.all(ds.sw_in.sel(time=slice("2019-10-06 00:00", "2019-10-06 03:00")).notnull())

    config.meteo.interpolation.cloudiness.allow_fallback = True
    model = oa.OpenAmundsen(config)
    model.initialize()
    model.meteo.sw_in.loc[:, nan_pos] = np.nan
    model.run()
    ds = model.point_output.data
    assert np.all(ds.sw_in.notnull())
    # Daytime values should be calculated using the "humidity" method here (not constant)
    assert not np.array_equal(
        ds.cloud_factor.loc["2019-10-06 12:00"].values,
        ds.cloud_factor.loc["2019-10-06 15:00"].values,
    )

    config.meteo.interpolation.cloudiness.clear_sky_fraction_night_method = "humidity"
    model = oa.OpenAmundsen(config)
    model.initialize()
    model.run()
    ds = model.point_output.data
    # Nighttime values should not be constant here
    assert not np.array_equal(
        ds.cloud_factor.loc["2019-10-06 00:00"].values,
        ds.cloud_factor.loc["2019-10-06 03:00"].values,
    )

    config.meteo.interpolation.cloudiness.clear_sky_fraction_night_method = "constant"
    model = oa.OpenAmundsen(config)
    model.initialize()
    model.run()
    ds = model.point_output.data
    # Nighttime values should be constant here
    assert_equal(
        ds.cloud_factor.loc["2019-10-06 00:00"].values,
        ds.cloud_factor.loc["2019-10-06 03:00"].values,
    )


def test_humidity():
    nan_pos = slice("2019-10-06 00:00", "2019-10-07 00:00")

    config = base_cloudiness_config()
    config.meteo.interpolation.cloudiness.method = "humidity"
    config.meteo.interpolation.cloudiness.allow_fallback = False
    model = oa.OpenAmundsen(config)
    model.initialize()

    model.meteo.rel_hum.loc["proviantdepot", nan_pos] = np.nan
    model.run()
    ds = model.point_output.data
    assert np.all(ds.sw_in.notnull())

    model = oa.OpenAmundsen(config)
    model.initialize()
    model.meteo.rel_hum.loc[:, nan_pos] = np.nan
    model.run()
    ds = model.point_output.data
    assert np.all(ds.sw_in.sel(time=nan_pos).isnull())
    assert np.all(ds.sw_in.drop_sel(time=ds.sel(time=nan_pos).time).notnull())


def test_prescribed(tmp_path):
    config = base_cloudiness_config()
    config.meteo.interpolation.cloudiness.method = "clear_sky_fraction"
    config.meteo.interpolation.cloudiness.allow_fallback = False
    model = oa.OpenAmundsen(config)
    model.initialize()
    model.run()
    ds_ref = model.point_output.data

    for station_id in ds_ref.indexes["point"]:
        ds = xr.load_dataset(f"{config.input_data.meteo.dir}/{station_id}.nc")
        ds = ds.sel(time=ds_ref.time)
        ds["cloud_fraction"] = xr.DataArray(
            ds_ref.cloud_fraction.sel(point=station_id).values * 100,
            coords=ds.coords,
            dims=ds.dims,
        )
        ds = ds.drop_vars("sw_in", errors="ignore")
        ds.to_netcdf(f"{tmp_path}/{station_id}.nc")

    config.meteo.interpolation.cloudiness.method = "prescribed"
    model = oa.OpenAmundsen(config)
    with pytest.raises(errors.MeteoDataError):
        model.initialize()

    config.input_data.meteo.dir = str(tmp_path)
    model = oa.OpenAmundsen(config)
    model.initialize()
    model.run()
    ds = model.point_output.data
    assert_allclose(ds_ref.cloud_fraction, ds.cloud_fraction, atol=0.09)
    # (differences are (likely) due to interpolation from station positions to grid points)

    # Test if clear_sky_fraction is used when allow_fallback = True and cloudiness values are nan
    config.meteo.interpolation.cloudiness.allow_fallback = True
    config.meteo.interpolation.cloudiness.method = "clear_sky_fraction"
    model = oa.OpenAmundsen(config)
    model.initialize()
    model.run()
    ds_csf = model.point_output.data
    config.meteo.interpolation.cloudiness.method = "prescribed"
    model = oa.OpenAmundsen(config)
    model.initialize()
    model.meteo["cloud_fraction"].loc[:] = np.nan
    model.run()
    ds = model.point_output.data
    assert_equal(ds_csf["cloud_fraction"].values, ds["cloud_fraction"].values)
    # (use ds_csf instead of ds_ref because for ds_ref allow_fallback is False and hence the
    # results differ for time steps with missing radiation data)


def test_outside_grid_data(tmp_path):
    config = base_cloudiness_config()
    config.meteo.interpolation.cloudiness.allow_fallback = False
    config.output_data.timeseries.points.append(
        {
            "x": 638483,
            "y": 5190972,
        }
    )
    config.output_data.timeseries.add_default_points = False

    config.meteo.interpolation.cloudiness.method = "clear_sky_fraction"
    model = oa.OpenAmundsen(config)
    model.initialize()
    model.run()
    ds_ref_csf = model.point_output.data

    config.meteo.interpolation.cloudiness.method = "humidity"
    model = oa.OpenAmundsen(config)
    model.initialize()
    model.run()
    ds_ref_hum = model.point_output.data

    src_dir = Path(config.input_data.grids.dir)
    dst_dir = tmp_path
    window = rasterio.windows.Window(col_off=5, row_off=2, width=5, height=4)
    src_files = src_dir.glob("*.asc")
    for src_file in src_files:
        with rasterio.open(src_file) as src_ds:
            kwargs = src_ds.meta.copy()
            kwargs.update(
                {
                    "height": window.height,
                    "width": window.width,
                    "transform": rasterio.windows.transform(window, src_ds.transform),
                }
            )

            with rasterio.open(dst_dir / src_file.name, "w", **kwargs) as dst_ds:
                dst_ds.write(src_ds.read(window=window))

    config.input_data.grids.dir = str(dst_dir)

    config.meteo.interpolation.cloudiness.method = "clear_sky_fraction"
    model = oa.OpenAmundsen(config)
    model.initialize()
    assert not np.any(model.meteo.within_grid_extent)
    model.run()
    ds = model.point_output.data
    # Daytime values must be nan
    assert np.all(ds.sw_in.sel(time=ds.indexes["time"].hour.isin([9, 12, 15])).isnull())

    config.meteo.interpolation.cloudiness.method = "humidity"
    model = oa.OpenAmundsen(config)
    model.initialize()
    model.run()
    ds = model.point_output.data
    assert np.all(ds.sw_in.notnull())
    assert_allclose(ds_ref_hum.sw_in, ds.sw_in, atol=25)

    # Now enable the extended grids
    grid_file_suffix = f"{config.domain}_{config.resolution}.asc"
    (tmp_path / f"extended-dem_{grid_file_suffix}").symlink_to(src_dir / f"dem_{grid_file_suffix}")
    (tmp_path / f"extended-svf_{grid_file_suffix}").symlink_to(src_dir / f"svf_{grid_file_suffix}")

    config.meteo.interpolation.cloudiness.method = "clear_sky_fraction"
    model = oa.OpenAmundsen(config)
    model.initialize()
    model.run()
    ds = model.point_output.data
    assert np.all(ds.sw_in.notnull())
    assert_allclose(ds_ref_csf.sw_in, ds.sw_in, rtol=0.03)

    config.meteo.interpolation.cloudiness.method = "humidity"
    model = oa.OpenAmundsen(config)
    model.initialize()
    model.run()
    ds = model.point_output.data
    assert np.all(ds.sw_in.notnull())
    assert_allclose(ds_ref_hum.sw_in, ds.sw_in, rtol=0.04)



from pathlib import Path
from typing import Dict, List

import numpy as np
import pandas as pd
import xarray as xr
from loguru import logger

from openamundsen import constants, errors, forcing, util
from openamundsen import meteo as oameteo


def read_meteo_data(
    meteo_format,
    meteo_data_dir,
    start_date,
    end_date,
    meteo_crs=None,
    grid_crs=None,
    bounds=None,
    exclude=None,
    include=None,
    filters=None,
    freq="h",
    aggregate=False,
):
    """
    Read all available stations in NetCDF or CSV format for a given period.

    In case of NetCDF input, all available .nc files in `meteo_data_dir` are
    used.
    For CSV input a file named `stations.csv` must be provided in
    the specified directory containing the metadata (IDs, names, x/y
    coordinates, altitudes) of the stations. Station files must be in the same
    directory and must be named `<station_id>.csv`

    Parameters
    ----------
    meteo_format : str
        Data format (either 'netcdf' or 'csv').

    meteo_data_dir : path-like
        Location of the NetCDF files.

    start_date : datetime-like
        Start date.

    end_date : datetime-like
        End date.

    meteo_crs : str
        CRS of the station coordinates specified in the stations.csv file
        (required only when `meteo_format` is 'csv').

    grid_crs : str
        CRS of the model grid.

    bounds : list or None
        If specified, use only stations within the region specified as a list of
        (x_min, y_min, x_max, y_max) coordinates in the model grid CRS.

    exclude: list or None
        List of station IDs to exclude.

    include: list or None
        List of station IDs to include even if otherwise excluded via `bounds`
        or `exclude`.

    freq : str
        Pandas-compatible frequency string (e.g. '3h') to which the data should
        be resampled.

    aggregate : boolean, default False
        Aggregate data when downsampling to a lower frequency or take
        instantaneous values.

    Returns
    -------
    ds : Dataset
        Station data.
    """
    meteo_data_dir = Path(meteo_data_dir)

    meta = _read_meteo_metadata(meteo_format, meteo_data_dir, meteo_crs, grid_crs)
    if meta.empty:
        raise errors.MeteoDataError("No stations found")

    meta = _apply_station_rules(meta, bounds, exclude, include)
    if meta.empty:
        raise errors.MeteoDataError("No stations available after applying station rules")

    datasets = []
    for station_id in meta.index:
        if meteo_format == "netcdf":
            filename = meteo_data_dir / f"{station_id}.nc"
            logger.info(f"Reading meteo file: {filename}")
            ds = read_netcdf_meteo_file(filename)
        elif meteo_format == "csv":
            filename = meteo_data_dir / f"{station_id}.csv"
            logger.info(f"Reading meteo file: {filename}")
            ds = read_csv_meteo_file(
                filename,
                station_id,
                meta.loc[station_id, "name"],
                meta.loc[station_id, "x"],
                meta.loc[station_id, "y"],
                meta.loc[station_id, "alt"],
                grid_crs,
            )

        if filters is not None:
            ds = _apply_filters(ds, filters)

        ds = _slice_and_resample_dataset(
            ds,
            start_date,
            end_date,
            freq,
            aggregate=aggregate,
        )

        if ds.sizes["time"] == 0:
            logger.warning("File contains no meteo data for the specified period")
        else:
            datasets.append(ds)

    if len(datasets) == 0:
        raise errors.MeteoDataError("No meteo data available for the specified period")

    ds_combined = forcing.combine_point_datasets(datasets)
    dates_combined = ds_combined.time.to_index()
    if dates_combined[0] > start_date or dates_combined[-1] < end_date:
        raise errors.MeteoDataError(
            "Insufficient meteo data available.\n"
            f"Requested period: {start_date}..{end_date}\n"
            f"Available period: {dates_combined[0]}..{dates_combined[-1]}"
        )

    return ds_combined


def read_netcdf_meteo_file(filename):
    """
    Read a meteo data file in NetCDF format and
    - check if the time, lon, lat, and alt variables are included
    - rename the variables according to NETCDF_VAR_MAPPINGS
    - convert precipitation rates to sums if necessary
    - check if the units are as expected (METEO_VAR_METADATA)
    - remove all unsupported variables
    - set the station id and name as attributes.

    Parameters
    ----------
    filename : path-like
        CSV filename.

    Returns
    -------
    ds : Dataset
        Station data.
    """
    ds = xr.load_dataset(filename)

    # rename variables
    ds_vars = list(ds.variables.keys())
    rename_vars = set(constants.NETCDF_VAR_MAPPINGS.keys()) & set(ds_vars)
    rename_dict = {v: constants.NETCDF_VAR_MAPPINGS[v] for v in rename_vars}
    ds = ds.rename_vars(rename_dict)

    ds = forcing.make_point_dataset(data=ds, point_id=_netcdf_station_id(filename))
    return ds


def read_csv_meteo_file(filename, station_id, station_name, x, y, alt, crs):
    """
    Read a meteorological data file in CSV format and return it as a Dataset.
    Unlike in read_netcdf_meteo_file, here it is assumed that precipitation is
    specified as a sum over the time step (i.e., kg m-2) instead of a rate (kg
    m-2 s-1).

    Parameters
    ----------
    filename : path-like
        CSV filename.

    station_id : str
        Station ID (must be unique).

    station_name : str
        Station name.

    x : float
        Station x coordinate.

    y : float
        Station y coordinate.

    alt : float
        Station altitude.

    crs : str
        CRS of the x/y coordinates.

    Returns
    -------
    ds : Dataset
        Station data.
    """
    lon, lat = util.transform_coords(x, y, crs, constants.CRS_WGS84)
    df = pd.read_csv(filename, parse_dates=True, index_col=0)

    if "precip" in df.columns:
        pass
        # TODO check if the time step is really 3600 s and if precip is really a sum

    ds = forcing.make_point_dataset(
        data=df,
        point_id=station_id,
        name=station_name,
        lon=lon,
        lat=lat,
        alt=alt,
    )

    return ds


def _apply_filters(ds: xr.Dataset, filters: List[Dict]) -> xr.Dataset:
    if len(filters) == 0:
        return ds

    ds = ds.copy()

    for filter_dict in filters:
        var = filter_dict["var"]
        if var in ds:
            ds[var] = _apply_filter(ds[var], filter_dict)

    return ds


def _apply_filter(data: xr.DataArray, f: Dict) -> xr.DataArray:
    data = data.copy(deep=True)

    if f["filter"] == "range":
        min_range = f.get("min", np.nan)
        max_range = f.get("max", np.nan)

        if np.isfinite(min_range):
            data[data < min_range] = np.nan

        if np.isfinite(max_range):
            data[data > max_range] = np.nan
    else:
        raise NotImplementedError

    return data


def _slice_and_resample_dataset(ds, start_date, end_date, freq, aggregate=False):
    """
    Slice a dataset to a given date range and optionally resample to a given
    time frequency.

    Parameters
    ----------
    ds : xr.Dataset
        Dataset.

    start_date : datetime-like
        Start date.

    end_date : datetime-like
        End date.

    freq : str
        Pandas-compatible frequency string (e.g. '3h'). Must be an exact subset
        of the original frequency of the data.

    aggregate : boolean, default False
        Aggregate data when downsampling to a lower frequency or take
        instantaneous values.

    Returns
    -------
    ds : Dataset
        Sliced and resampled dataset.
    """
    freq_td = util.offset_to_timedelta(freq)
    inferred_freq = ds.time.to_index().inferred_freq
    td_1d = pd.Timedelta("1d")

    if inferred_freq is None and ds.sizes["time"] > 2:
        # ("> 2" because inferring the frequency requires at least 3 points, so for shorter
        # time series inferred_freq is always None)
        raise errors.MeteoDataError("File contains missing records or non-uniform timesteps")

    if freq_td == td_1d and util.offset_to_timedelta(inferred_freq) < td_1d:
        if not aggregate:
            logger.warning(
                "File contains sub-daily data. Data will be aggregated to daily time steps despite "
                "aggregate_when_downsampling being set to False."
            )
            aggregate = True

        end_date = (end_date + td_1d).normalize()

    if inferred_freq is not None and inferred_freq != freq:
        # keep data from previous day for aggregation, if present
        ds = ds.sel(time=slice(start_date - td_1d, end_date))
        ds = _resample_dataset(ds, start_date, end_date, freq, aggregate=aggregate)

    ds = ds.sel(time=slice(start_date, end_date))

    return ds


def _resample_dataset(ds, start_date, end_date, freq, aggregate=False):
    """
    Resample a dataset to a given time frequency.

    Parameters
    ----------
    ds : xr.Dataset
        Dataset.

    start_date : datetime-like
        Start date.

    end_date : datetime-like
        End date.

    freq : str
        Pandas-compatible frequency string (e.g. '3h'). Must be an exact subset
        of the original frequency of the data.

    aggregate : boolean, default False
        Aggregate data when downsampling to a lower frequency or take
        instantaneous values.

    Returns
    -------
    ds_res : Dataset
        Resampled dataset.
    """
    td = util.offset_to_timedelta(freq)
    td_1d = pd.Timedelta("1d")
    if td < td_1d:
        resample_kwargs = {
            "label": "right",
            "closed": "right",
            "origin": pd.Timestamp(start_date),
        }
    elif td == td_1d:
        resample_kwargs = {
            "label": "left",
            "closed": "right",
            "origin": "start",
        }
    else:
        raise errors.MeteoDataError("Resampling to frequencies > 1 day is not supported")

    if ds.sizes["time"] == 0:
        # Nothing to resample
        return ds

    # ds.resample() is extremely slow for some reason, so we resample using pandas
    df = ds.to_dataframe().drop(columns=["lon", "lat", "alt"])

    if aggregate:
        # Calculate averages
        df_res = df.resample(freq, **resample_kwargs).mean()

        # Wind direction must be aggregated separately
        if "wind_dir" in df.columns:
            df_res["wind_dir"] = _aggregate_wind_dir(df, freq, resample_kwargs)

        # We might end up with an extra bin after resampling; remove it here
        df_res = df_res.loc[start_date:end_date]
        if df_res.index[0] < df.index[0]:
            df_res = df_res.iloc[1:]
    else:
        # Take the instantaneous values
        df_res = df.reindex(
            pd.date_range(
                start=start_date,
                end=end_date,
                freq=freq,
                name="time",
            )
        ).loc[df.index[0] : df.index[-1]]

    # Precipitation is summed up regardless of the aggregation setting
    if "precip" in df:
        df_res["precip"] = (
            df["precip"]
            .resample(
                freq,
                **resample_kwargs,
            )
            .agg(pd.Series.sum, skipna=False)
            .reindex(df_res.index)  # to avoid potentially inserting new rows into df_res
        )

    # For wind gusts always take the maximum value
    if "wind_speed_gust" in df.columns:
        df_res["wind_speed_gust"] = (
            df["wind_speed_gust"]
            .resample(
                freq,
                **resample_kwargs,
            )
            .max()
            .reindex(df_res.index)  # to avoid potentially inserting new rows into df_res
        )

    # Check if the desired frequency is a subset of the original frequency of the
    # data (e.g., resampling hourly to 3-hourly data is ok, but not hourly to
    # 1.5-hourly, or upsampling in general)
    if not df.index.intersection(df_res.index).equals(df_res.index):
        raise errors.MeteoDataError(
            f'Resampling from freq "{df.index.inferred_freq}" to "{freq}" not supported'
        )

    # Create a new dataset with the resampled time series
    ds_res = ds[["lon", "lat", "alt", "time"]]
    ds_res["time"] = df_res.index
    ds_res.attrs = ds.attrs

    for param in df_res.columns:
        ds_res[param] = df_res[param]

    return ds_res


def _netcdf_station_id(filename):
    """
    Return the station ID of a station in NetCDF format, i.e., the base name of
    the file without the .nc extension.
    """
    return Path(filename).stem


def _read_meteo_metadata(meteo_format, meteo_data_dir, meteo_crs, grid_crs):
    """
    Read the metadata of the available meteorological stations.

    Parameters
    ----------
    meteo_format : str
        Data format (either 'netcdf' or 'csv').

    meteo_data_dir : path-like
        Location of the NetCDF files.

    meteo_crs : str
        CRS of the station coordinates specified in the stations.csv file
        (required only when `meteo_format` is 'csv').

    grid_crs : str
        CRS of the model grid.

    Returns
    -------
    meta : DataFrame
        DataFrame containing the station IDs as index and the columns `name`,
        `x` (x coordinate in the grid CRS), `y` (y coordinate in the grid CRS),
        and `alt`.
    """
    if meteo_format == "netcdf":
        nc_files = sorted(meteo_data_dir.glob("*.nc"))
        if len(nc_files) == 0:
            raise errors.MeteoDataError(f"No meteo data files found in {meteo_data_dir}")

        station_ids = []
        names = []
        lons = []
        lats = []
        alts = []

        for nc_file in nc_files:
            with xr.open_dataset(nc_file) as ds:
                station_ids.append(_netcdf_station_id(nc_file))
                names.append(None)  # name is not needed here, is set in read_netcdf_meteo_file
                lons.append(float(ds.lon.values))
                lats.append(float(ds.lat.values))
                alts.append(float(ds.alt.values))

        # Transform lon/lat coordinates to grid CRS
        xs_gridcrs, ys_gridcrs = util.transform_coords(
            lons,
            lats,
            constants.CRS_WGS84,
            grid_crs,
        )

        meta = pd.DataFrame(
            index=station_ids,
            data={
                "name": names,
                "x": xs_gridcrs,
                "y": ys_gridcrs,
                "alt": alts,
            },
        )
    elif meteo_format == "csv":
        filename = f"{meteo_data_dir}/stations.csv"
        try:
            meta = pd.read_csv(filename, index_col=0)
        except FileNotFoundError as err:
            raise errors.MeteoDataError(f"Missing station metadata file ({filename})") from err

        required_cols = ["name", "x", "y", "alt"]
        if not all(pd.Index(required_cols).isin(meta.columns)):
            raise errors.MeteoDataError(
                f"Station metadata file does not contain all required columns ({required_cols})"
            )

        # Transform x/y coordinates from meteo CRS to grid CRS
        xs_gridcrs, ys_gridcrs = util.transform_coords(
            meta.x,
            meta.y,
            meteo_crs,
            grid_crs,
        )

        meta.x = xs_gridcrs
        meta.y = ys_gridcrs
    else:
        raise NotImplementedError

    return meta


def _apply_station_rules(meta, bounds, exclude, include):
    """
    Extend or reduce the list of stations to read in using a defined region or
    exclusion/inclusion lists.

    Parameters
    ----------
    meta : DataFrame
        Station metadata as returned by `_read_meteo_metadata`.

    bounds : list or None
        If specified, use only stations within the region specified as a list of
        (x_min, y_min, x_max, y_max) coordinates in the model grid CRS.

    exclude: list or None
        List of station IDs to exclude.

    include: list or None
        List of station IDs to include even if otherwise excluded via `bounds`
        or `exclude`.

    Returns
    -------
    meta : DataFrame
    """
    meta_all = meta.copy()

    if bounds is not None:
        # TODO check if using ">=" and "<" is correct here or if ">" and "<=" should be used for y
        # coordinates
        meta = meta[
            (meta.x >= bounds[0])
            & (meta.y >= bounds[1])
            & (meta.x < bounds[2])
            & (meta.y < bounds[3])
        ]

    if exclude is not None:
        meta = meta.drop(index=exclude, errors="ignore")

    if include is not None:
        for station_id in include:
            try:
                meta = pd.concat([meta, meta_all.loc[[station_id]]])
            except KeyError as err:
                raise errors.MeteoDataError(
                    f"Station to be included not found: {station_id}"
                ) from err

    meta = meta[~meta.index.duplicated(keep="first")]  # remove potential duplicate entries
    return meta


def _aggregate_wind_dir(df, freq, resample_kwargs):
    """
    Aggregate wind direction by averaging the wind vector components.

    Parameters
    ----------
    df : DataFrame
        Original data. Must contain a "wind_dir" column.

    freq : str
        Aggregation frequency.

    resample_kwargs : dict
        Keyword arguments to be passed to resample().

    Returns
    -------
    wind_dir : Series
        Aggregated wind direction.
    """
    df = df.copy()

    try:
        ws = df["wind_speed"]
    except KeyError:
        ws = df["wind_dir"] * float("nan")

    wind_us, wind_vs = oameteo.wind_to_uv(ws, df["wind_dir"])
    df["wind_u"] = wind_us
    df["wind_v"] = wind_vs
    df = df[["wind_u", "wind_v"]]

    df_res = df.resample(freq, **resample_kwargs).mean()
    _, wd_res = oameteo.wind_from_uv(df_res["wind_u"], df_res["wind_v"])
    df_res["wind_dir"] = wd_res

    return df_res["wind_dir"]



import importlib.util
import textwrap
from pathlib import Path

import pooch
import pytest

import openamundsen as oa

version = oa.__version__
if "+" in version:
    data_url = "https://github.com/openamundsen/testdata/raw/{version}/"
else:
    data_url = "https://github.com/openamundsen/testdata/raw/v{version}/"

data_fetcher = pooch.create(
    path=pooch.os_cache("openamundsen-tests"),
    base_url=data_url,
    version=version,
    version_dev="main",
    registry={  # hashes generated with generate_hashes.py
        "grids/rofental/dem_rofental_1000.asc": "368ee374bfbb927bae5408a38dbafbbeab4ff59dc94ae30956b617bf28c114ec",  # noqa: E501
        "grids/rofental/lc_rofental_1000.asc": "5bde204bf9e79acdc47f9d9c70f21f20f669cfb7cbe65418ef42dd0a52676b98",  # noqa: E501
        "grids/rofental/roi_rofental_1000.asc": "5c495833e8465931d7aedc982ef093332572034d9cf7da5fc8b98b72111e3d25",  # noqa: E501
        "grids/rofental/soil_rofental_1000.asc": "48b46a8b3216a012393f36e34561473b3c1fd8dccef86d9a7e34aab9e3deb89b",  # noqa: E501
        "grids/rofental/srf_rofental_1000.asc": "431f0ed2181abc2a79c6ffd79b804e5c983c9c47003bce57b853db2b41abf62d",  # noqa: E501
        "grids/rofental/svf_rofental_1000.asc": "a4b7704fad6f7c63445a76c64207ff881b73023ec7d958d4581cd7fb42d06966",  # noqa: E501
        "meteo/rofental/convert_csv_to_netcdf.py": "f4e7aeb04d3eff7f1bdc6c4619600c4ab8dc560ffbc8feab83a30d200400f49f",  # noqa: E501
        "meteo/rofental/csv/bellavista.csv": "0c760f490aa7dc8d6af9da8987a6cf59807ec3e41c802a9c73b7773d269ec603",  # noqa: E501
        "meteo/rofental/csv/latschbloder.csv": "f9e089bca0e7c931df51e1c624dc4985089c20c802d43de8b4c5175353a00931",  # noqa: E501
        "meteo/rofental/csv/proviantdepot.csv": "0e3a7eb07f9f127f0a257b17c62442e91860fb5a64f57387da547f69dcb61b6f",  # noqa: E501
        "meteo/rofental/csv/stations.csv": "175b420d9659d3478d4755fc4b8d8c2569400b37d84ffb9906e1ab661afe5dc4",  # noqa: E501
        "meteo/rofental/netcdf/bellavista.nc": "0fdeb860bce28a287bd48d2fc95f42a07ed842e9a0f0a0045f658605df2ff564",  # noqa: E501
        "meteo/rofental/netcdf/latschbloder.nc": "cc96e0dbc4024b093fcc161d8f2c862bb8abd379a7d3dbd540d78242f17aef75",  # noqa: E501
        "meteo/rofental/netcdf/proviantdepot.nc": "f8a276c98343c2574fb810eb4b3ee1b08bd29d9f5d0cfaa5a1104f1aa25af374",  # noqa: E501
        "results/canopy_point.nc": "3cd25b0119037be0eb95401ae7cfb618fc840d5244e4b2f42af45e72feff493c",  # noqa: E501
        "results/interpolation_hum_fixed.nc": "ca56dcefe7e832745c2fd17686168b0a6cc5ad57d3cf29bb5ab6f803e0126185",  # noqa: E501
        "results/interpolation_hum_piecewise-regression.nc": "28024bfa104b5d58ae5251002f64035520becbfe4834f5b97ef3386eddf75e48",  # noqa: E501
        "results/interpolation_hum_regression.nc": "28024bfa104b5d58ae5251002f64035520becbfe4834f5b97ef3386eddf75e48",  # noqa: E501
        "results/interpolation_precip_adjustment_factor.nc": "1aefa30dfd93d3cd4cc68fc0efea1168e069f0190fa7b9a5de123352aff182c2",  # noqa: E501
        "results/interpolation_precip_fractional.nc": "c7b0f660454e3cdffe30c62702ffc9e6f95ba997455c4fd130e08caa3368690b",  # noqa: E501
        "results/interpolation_precip_regression.nc": "a9e7db0df880946e3731ad92da835b85c6819051c22c3735c8ca585e6f38d8a6",  # noqa: E501
        "results/interpolation_temp_fixed.nc": "e20b3a01827c0fdb9c5c465dbd1145a77bfe5744edd52ac6bd8355f9caacbf78",  # noqa: E501
        "results/interpolation_temp_piecewise-regression.nc": "9ace6b08730b7c9be29c4723046dc5dc82d0c723faab374f260c3f6276b48fb5",  # noqa: E501
        "results/interpolation_temp_regression.nc": "9ace6b08730b7c9be29c4723046dc5dc82d0c723faab374f260c3f6276b48fb5",  # noqa: E501
        "results/interpolation_wind_regression.nc": "52e84142f6a4b9ec9fd7c591741a61d52898afea6d0ffe8675e1c168c30fe10f",  # noqa: E501
        "results/snow_cryolayers_point.nc": "04c6363c888bd6296e9ed3fa5325e418c659532b4f066d24fd33a34d5930877d",  # noqa: E501
        "results/snow_multilayer_point.nc": "702b7879a57f9837c97a629e66cd11fbb014d1b6e42f1241874de86d0714c380",  # noqa: E501
    },
)
DATA_DIR = data_fetcher.abspath

_BASE_CONFIG_YAML = textwrap.dedent(f"""
    domain: rofental
    start_date: 2020-01-15
    end_date: 2020-01-15
    resolution: 1000
    timestep: 3h
    crs: "epsg:32632"
    timezone: 1

    input_data:
      grids:
        dir: {DATA_DIR}/grids/rofental
      meteo:
        dir: {DATA_DIR}/meteo/rofental/netcdf

    output_data:
      timeseries:
        format: memory

      grids:
        format: memory
""")
_BASE_CONFIG = oa.parse_config(oa.Configuration.from_yaml(_BASE_CONFIG_YAML))


def pytest_addoption(parser):
    parser.addoption("--run-slow", action="store_true", default=False, help="run slow tests")
    parser.addoption(
        "--skip-comparisons",
        action="store_true",
        default=False,
        help="skip comparisons",
    )
    parser.addoption(
        "--prepare-comparison-data",
        action="store_true",
        default=False,
        help="prepare baseline model results for comparisons",
    )
    parser.addoption(
        "--comparison-data-dir",
        type=str,
        help="baseline model results directory",
    )
    parser.addoption(
        "--reports-dir",
        type=str,
        help="directory for writing comparison reports",
    )


def pytest_configure(config):
    config.addinivalue_line("markers", "slow: marks test as slow")
    config.addinivalue_line("markers", "comparison: marks test as comparing with baseline data")

    prepare_comparison_data = config.getoption("--prepare-comparison-data")

    comp_data_dir = config.getoption("--comparison-data-dir")
    if comp_data_dir is None:
        if prepare_comparison_data:
            raise ValueError(
                "--comparison-data-dir must be specified when --prepare-comparison-data is set"
            )
        else:
            comp_data_dir = DATA_DIR / "results"
    if comp_data_dir is not None:
        comp_data_dir = Path(comp_data_dir)
        comp_data_dir.mkdir(parents=True, exist_ok=True)

    reports_dir = config.getoption("--reports-dir")
    if reports_dir is not None:
        reports_dir = Path(reports_dir)

        if importlib.util.find_spec("plotly") is None:
            raise ImportError("plotly is required for report creation")

    pytest.DATA_DIR = DATA_DIR
    pytest.COMPARISON_DATA_DIR = comp_data_dir
    pytest.REPORTS_DIR = reports_dir
    pytest.PREPARE_COMPARISON_DATA = prepare_comparison_data


def pytest_collection_modifyitems(config, items):
    run_slow = config.getoption("--run-slow")
    skip_slow = pytest.mark.skip(reason="need --run-slow option to run")
    skip_comparisons = pytest.mark.skipif(
        config.getoption("--skip-comparisons"),
        reason="skipping comparisons",
    )

    for item in items:
        if "slow" in item.keywords and not run_slow:
            item.add_marker(skip_slow)
        elif "comparison" in item.keywords:
            item.add_marker(skip_comparisons)


def pytest_sessionstart(session):
    fetch_data_files()


def fetch_data_files():
    for file in data_fetcher.registry_files:
        data_fetcher.fetch(file)


def base_config():
    return _BASE_CONFIG.copy()



class SnowModel:
    def initialize(self):
        pass

    def albedo_aging(self):
        pass

    def compaction(self):
        pass

    def accumulation(self):
        pass

    def heat_conduction(self):
        pass

    def melt(self):
        pass

    def sublimation(self):
        pass

    def runoff(self):
        pass

    def update_layers(self):
        pass

    def update_properties(self):
        pass

    def add_snow(
        self,
        pos,
        ice_content,
        liquid_water_content=0,
        density=None,
        albedo=None,
    ):
        pass



from . import constants, errors, terrain
from .conf import Configuration, parse_config, read_config
from .model import Model, OpenAmundsen


# Get version (method as used by matplotlib: https://github.com/matplotlib/matplotlib/blob/bcc1ce8461f5b6e874baaaa02ef776d0243a4abe/lib/matplotlib/__init__.py#L133-L151)
def __getattr__(name):
    if name == "__version__":
        from pathlib import Path

        import setuptools_scm

        global __version__
        root = Path(__file__).resolve().parents[1]
        if (root / ".git").exists() and not (root / ".git/shallow").exists():
            __version__ = setuptools_scm.get_version(
                root=root,
                version_scheme="post-release",
                fallback_version="0.0.0+UNKNOWN",
            )
        else:
            try:
                from . import _version

                __version__ = _version.version
            except ImportError:
                __version__ = "0.0.0+UNKNOWN"
        return __version__
    raise AttributeError(f"module {__name__!r} has no attribute {name!r}")


__all__ = [
    "Configuration",
    "Model",
    "OpenAmundsen",
    "constants",
    "errors",
    "parse_config",
    "read_config",
    "terrain",
]



import tempfile
from copy import deepcopy
from textwrap import dedent

import pandas as pd
import pytest
from ruamel.yaml import YAML

import openamundsen as oa
import openamundsen.errors as errors


def save_and_read_yaml_dict(d):
    yaml = YAML()

    with tempfile.TemporaryDirectory() as temp_dir:
        filename = f"{temp_dir}/config.yml"
        with open(filename, "w") as f:
            yaml.dump(d, f)

        return oa.read_config(filename)


def parse_yaml_config(config):
    return oa.parse_config(oa.Configuration.from_yaml(config))


def save_and_parse_config(config):
    with tempfile.TemporaryDirectory() as temp_dir:
        filename = f"{temp_dir}/config.yml"

        with open(filename, "w") as f:
            if isinstance(config, dict):
                yaml = YAML()
                yaml.dump(config, f)
            elif isinstance(config, str):
                f.write(config)
                f.flush()
            else:
                raise TypeError("Unsupported type")

        return oa.parse_config(oa.read_config(filename))


@pytest.fixture(scope="function")
def minimal_config():
    return {
        "domain": "dummy",
        "start_date": "2019-11-01",
        "end_date": "2020-04-30",
        "crs": "epsg:32632",
        "resolution": 50,
        "timezone": 1,
    }


def test_config_equal():
    config = {
        "domain": "dummy",
        "start_date": "2019-11-01",
        "end_date": "2020-04-30",
        "crs": "epsg:32632",
        "resolution": 50,
        "timezone": 1,
        "input_data": {
            "grids": {
                "dir": "dummydir",
            },
            "meteo": {
                "dir": "meteodummydir",
                "format": "netcdf",
            },
        },
    }

    config_read = save_and_read_yaml_dict(config)

    assert config == config_read


def test_read_config(minimal_config):
    config = save_and_parse_config(minimal_config)
    assert config["end_date"].hour == 23

    yaml_str = dedent("""
        domain: dummy
        start_date: 2019-11-01
        end_date: 2020-12-31
        crs: "epsg:32632"
        resolution: 50
        timezone: 1
    """)
    config = parse_yaml_config(yaml_str)
    assert config["end_date"].hour == 23


def test_start_end_date_order(minimal_config):
    mc = deepcopy(minimal_config)
    mc["start_date"] = "2020-01-01"
    mc["end_date"] = "2020-01-01"
    oa.parse_config(mc)
    mc["timestep"] = "D"
    oa.parse_config(mc)

    mc = deepcopy(minimal_config)
    mc["start_date"] = "2020-01-01 00:00"
    mc["end_date"] = "2020-01-01 00:00"
    oa.parse_config(mc)

    mc = deepcopy(minimal_config)
    mc["start_date"] = "2020-01-01"
    mc["end_date"] = "2019-12-31"
    with pytest.raises(errors.ConfigurationError):
        oa.parse_config(mc)

    mc = deepcopy(minimal_config)
    mc["start_date"] = "2020-01-01 01:00"
    mc["end_date"] = "2020-01-01 00:00"
    with pytest.raises(errors.ConfigurationError):
        oa.parse_config(mc)


def test_infer_end_date(minimal_config):
    config = oa.parse_config(minimal_config)
    assert config["end_date"].hour == 23

    mc = deepcopy(minimal_config)
    mc["end_date"] = "2020-04-30 11:00"
    config = oa.parse_config(mc)
    assert config["end_date"].hour == 11

    mc = deepcopy(minimal_config)
    mc["end_date"] = " 2020-04-30 "
    config = oa.parse_config(mc)
    assert config["end_date"].hour == 23

    mc = deepcopy(minimal_config)
    mc["timestep"] = "D"
    config = oa.parse_config(mc)
    assert config["end_date"].hour == 0

    mc = deepcopy(minimal_config)
    mc["start_date"] = "2017-11-01"
    mc["end_date"] = "2018-04-25"
    mc["timestep"] = "5D"
    mc["output_data"] = {
        "timeseries": {
            "write_freq": "30D",
        },
    }
    config = oa.parse_config(mc)
    assert config["end_date"] == pd.Timestamp(mc["end_date"])

    yaml_str = dedent("""
        domain: dummy
        start_date: 2019-11-01
        end_date: {d}
        crs: "epsg:32632"
        resolution: 50
        timezone: 1
    """)

    config = parse_yaml_config(yaml_str.format(d="2020-04-30"))
    assert config["end_date"].hour == 23

    config = parse_yaml_config(yaml_str.format(d='"2020-04-30"'))
    assert config["end_date"].hour == 23

    config = parse_yaml_config(yaml_str.format(d="2020-04-30 11:00"))
    assert config["end_date"].hour == 11

    config = parse_yaml_config(yaml_str.format(d='"2020-04-30 11:00"'))
    assert config["end_date"].hour == 11


def test_timesteps(minimal_config):
    mc = deepcopy(minimal_config)
    mc["start_date"] = "2019-11-01 03:00"
    oa.parse_config(mc)

    mc = deepcopy(minimal_config)
    mc["start_date"] = "2019-11-01 00:15"
    mc["end_date"] = "2019-11-01 23:45"
    mc["timestep"] = "15min"
    oa.parse_config(mc)

    mc = deepcopy(minimal_config)
    mc["start_date"] = "2019-11-01 02:00"
    mc["end_date"] = "2019-11-01 22:00"
    mc["timestep"] = "3h"
    with pytest.raises(errors.ConfigurationError):
        oa.parse_config(mc)

    mc = deepcopy(minimal_config)
    mc["end_date"] = "2019-12-31 15:00"
    mc["timestep"] = "D"
    with pytest.raises(errors.ConfigurationError):
        oa.parse_config(mc)

    mc = deepcopy(minimal_config)
    mc["start_date"] = "2019-12-31 01:23"
    mc["end_date"] = "2019-12-31 03:45"
    mc["timestep"] = "h"
    with pytest.raises(errors.ConfigurationError):
        oa.parse_config(mc)

    mc = deepcopy(minimal_config)
    mc["start_date"] = "2017-11-01"
    mc["end_date"] = "2018-04-29"
    mc["timestep"] = "5D"
    with pytest.raises(errors.ConfigurationError):
        oa.parse_config(mc)


def test_missing_parameter(minimal_config):
    for key in minimal_config.keys():
        config = deepcopy(minimal_config)
        del config[key]

        with pytest.raises(errors.ConfigurationError):
            oa.parse_config(config)


def test_parse_twice(minimal_config):
    c1 = oa.parse_config(minimal_config)
    c2 = oa.parse_config(c1)
    assert c1 == c2


def test_geographic_crs(minimal_config):
    minimal_config["crs"] = "epsg:4326"
    with pytest.raises(errors.ConfigurationError):
        oa.parse_config(minimal_config)



import numpy as np
from loguru import logger
from numba import njit, prange

from openamundsen.modules.radiation import shadows


def slope_aspect(dem, res):
    """
    Calculate terrain slope and aspect using Horn's formula.

    Parameters
    ----------
    dem : ndarray
        Terrain elevation (m).

    res : float
        DEM resolution (m).

    Returns
    -------
    slope : ndarray
        Slope (degrees).

    aspect : ndarray
        Aspect (degrees) in azimuth direction, i.e., 0° means North-facing, 90°
        East-facing, 180° South-facing, and 270° West-facing.
    """
    # TODO check/improve calculation for small DEMs
    if any(np.array(dem.shape) < (3, 3)):
        slope = np.full(dem.shape, 0)
        aspect = np.full(dem.shape, np.nan)
        return slope, aspect

    # 3x3 window around cell e is labeled as follows:
    # [ a b c ]
    # [ d e f ]
    # [ g h i ]

    za = np.roll(dem, [1, 1], axis=[0, 1])
    zb = np.roll(dem, 1, axis=0)
    zc = np.roll(dem, [1, -1], axis=[0, 1])
    zd = np.roll(dem, 1, axis=1)
    zf = np.roll(dem, -1, axis=1)
    zg = np.roll(dem, [-1, 1], axis=[0, 1])
    zh = np.roll(dem, -1, axis=0)
    zi = np.roll(dem, [-1, -1], axis=[0, 1])

    za[:, 0] = za[:, 1]
    za[0, :] = za[1, :]
    zb[0, :] = zb[1, :]
    zc[:, -1] = zc[:, -2]
    zc[0, :] = zc[1, :]
    zd[:, 0] = zd[:, 1]
    zf[:, -1] = zf[:, -2]
    zg[:, 0] = zg[:, 1]
    zg[-1, :] = zg[-2, :]
    zh[-1, :] = zh[-2, :]
    zi[-1, :] = zi[-2, :]
    zi[:, -1] = zi[:, -2]

    dzdx = ((zc + 2 * zf + zi) - (za + 2 * zd + zg)) / (8 * res)
    dzdy = ((zg + 2 * zh + zi) - (za + 2 * zb + zc)) / (8 * res)
    slope = np.rad2deg(np.arctan(np.sqrt(dzdx**2 + dzdy**2)))

    aspect = np.rad2deg(np.arctan2(dzdy, -dzdx))

    pos1 = aspect < 0
    pos2 = aspect > 90
    pos3 = (aspect >= 0) & (aspect <= 90)
    aspect[pos1] = 90 - aspect[pos1]
    aspect[pos2] = 360 - aspect[pos2] + 90
    aspect[pos3] = 90 - aspect[pos3]

    return slope, aspect


def normal_vector(dem, res):
    """
    Calculate a vector normal to the surface after Corripio (2003).

    Parameters
    ----------
    dem : ndarray
        Terrain elevation (m).

    res : float
        DEM resolution (m).

    Returns
    -------
    normal_vec : ndarray with dimensions (3, rows, cols)
        Unit vector perpendicular to the surface.

    References
    ----------
    .. [1] Corripio, J. G. (2003). Vectorial algebra algorithms for calculating
       terrain parameters from DEMs and solar radiation modelling in mountainous
       terrain. International Journal of Geographical Information Science, 17(1),
       1–23. https://doi.org/10.1080/13658810210157796
    """
    normal_vec = np.full((3, dem.shape[0], dem.shape[1]), np.nan)

    # TODO check/improve calculation for small DEMs
    # Here it is assumed that the DEM is flat
    if any(np.array(dem.shape) < (3, 3)):
        normal_vec[0, :, :] = 0
        normal_vec[1, :, :] = 0
        normal_vec[2, :, :] = 1
        return normal_vec

    right = np.roll(dem, -1, axis=1)
    up = np.roll(dem, 1, axis=0)
    diag = np.roll(dem, (1, -1), axis=(0, 1))

    normal_vec[0, :, :] = 0.5 * res * (dem - right + up - diag)
    normal_vec[1, :, :] = 0.5 * res * (dem + right - up - diag)
    normal_vec[2, :, :] = res**2

    # fill first row and last column
    normal_vec[:, 0, :] = normal_vec[:, 1, :]
    normal_vec[:, :, -1] = normal_vec[:, :, -2]

    normal_vec_len = np.sqrt(  # length of the normal vector
        normal_vec[0, :, :] ** 2 + normal_vec[1, :, :] ** 2 + normal_vec[2, :, :] ** 2
    )
    normal_vec /= normal_vec_len  # make it a unit vector

    return normal_vec


def sky_view_factor(
    dem,
    res,
    min_azim=0,
    max_azim=360,
    azim_step=10,
    elev_step=1,
    num_sweeps=1,
):
    """
    Calculate the sky view factor for a DEM after Corripio (2003).
    The sky view factor is the hemispherical fraction of unobstructed sky
    visible from any point. Calculation is performed using a hillshading
    algorithm at regular azimuth and elevation angle intervals.

    Parameters
    ----------
    dem : ndarray
        Terrain elevation (m).

    res : float
        DEM resolution (m).

    min_azim : int, default 0
        Minimum azimuth angle (degrees).

    max_azim : int, default 360
        Maximum azimuth angle (degrees).

    azim_step : int, default 10
        Azimuth angle interval (degrees).

    elev_step : int, default 1
        Elevation angle interval (degrees).

    num_sweeps : int, default 1
        Number of sweeps in each direction when calculating shadows.

    Returns
    -------
    svf : ndarray
        Sky view factor (values between 0 and 1).

    References
    ----------
    .. [1] Corripio, J. G. (2003). Vectorial algebra algorithms for calculating
       terrain parameters from DEMs and solar radiation modelling in mountainous
       terrain. International Journal of Geographical Information Science, 17(1),
       1–23. https://doi.org/10.1080/13658810210157796
    """
    # Set SVF to 1 for 1x1 grids
    if dem.shape == (1, 1):
        return np.full(dem.shape, 1.0)

    dem = dem.astype(float, copy=False)  # convert to float64 if necessary
    slope, _ = slope_aspect(dem, res)
    svf = np.zeros(dem.shape)

    azim_angles = np.arange(min_azim, max_azim, azim_step)
    min_elev_angle = 1
    max_elev_angle = max(
        int(np.ceil(np.nanmax(slope))),
        min_elev_angle,
    )
    elev_angles = np.arange(min_elev_angle, max_elev_angle + 1)[::-1]

    for azim_num, azim_angle in enumerate(azim_angles):
        logger.info(
            f"Calculating sky view factor for azimuth={azim_angle}° "
            f"({azim_num + 1}/{len(azim_angles)})"
        )

        azim_rad = np.deg2rad(azim_angle)
        svf_cur = np.full(dem.shape, max_elev_angle)

        for elev_angle in elev_angles:
            elev_rad = np.deg2rad(elev_angle)
            sun_vec = np.array(
                [
                    np.sin(azim_rad) * np.cos(elev_rad),
                    -np.cos(azim_rad) * np.cos(elev_rad),
                    np.sin(elev_rad),
                ]
            )

            illum = shadows(dem, res, sun_vec, num_sweeps=num_sweeps)
            svf_cur[illum == 1] = elev_angle

        svf += np.cos(np.deg2rad(svf_cur)) ** 2

    svf /= len(azim_angles)
    return svf


def curvature(dem, res, kind, L=None):
    """
    Calculate curvature for a digital terrain model.

    Parameters
    ----------
    dem : ndarray
        Terrain elevation (m).

    res : float
        DEM resolution (m).

    kind : str
        Curvature calculation method. Allowed values are:
            - 'liston': calculate topographic curvature according to [1].

    L : float, default None
        Length scale required for kind='liston' (m).

    Returns
    -------
    curv : ndarray
        Curvature array.

    References
    ----------
    .. [1] Liston, G. E., Haehnel, R. B., Sturm, M., Hiemstra, C. A.,
       Berezovskaya, S., & Tabler, R.  D. (2007). Simulating complex snow
       distributions in windy environments using SnowTran-3D. Journal of
       Glaciology, 53(181), 241–256.
    """
    if kind == "liston":
        if L is None:
            L = res

        L_px = int(np.ceil(L / res))

        z = dem
        z_n = _shift_arr(dem, 0, L_px, mode="edge")
        z_ne = _shift_arr(dem, 1, L_px, mode="edge")
        z_e = _shift_arr(dem, 2, L_px, mode="edge")
        z_se = _shift_arr(dem, 3, L_px, mode="edge")
        z_s = _shift_arr(dem, 4, L_px, mode="edge")
        z_sw = _shift_arr(dem, 5, L_px, mode="edge")
        z_w = _shift_arr(dem, 6, L_px, mode="edge")
        z_nw = _shift_arr(dem, 7, L_px, mode="edge")

        curv = (
            (z - (z_w + z_e) / 2.0) / (2 * L)
            + (z - (z_s + z_n) / 2.0) / (2 * L)
            + (z - (z_sw + z_ne) / 2.0) / (2 * np.sqrt(2) * L)
            + (z - (z_nw + z_se) / 2.0) / (2 * np.sqrt(2) * L)
        ) / 4.0
    else:
        raise NotImplementedError

    return curv


def openness(dem, res, L, negative=False, mean=True):
    """
    Calculate topographic openness for a DEM following [1].

    Parameters
    ----------
    dem : ndarray
        Terrain elevation (m).

    res : float
        DEM resolution (m).

    L : float
        Radial distance to consider for calculating openness for each pixel (m).

    negative : bool, default False
        Calculate negative instead of positive openness.

    mean : bool, default True
        Return openness averaged over all eight compass directions.

    Returns
    -------
    opn : ndarray
        Openness (radians).

    References
    ----------
    .. [1] Yokoyama, R., Shirasawa, M., & Pike, R. J. (2002). Visualizing
       topography by openness: A new application of image processing to digital
       elevation models. Photogrammetric Engineering and Remote Sensing, 68(3),
       257–266.
    """
    dirs = np.arange(8)
    opn = np.full((len(dirs), dem.shape[0], dem.shape[1]), np.inf)

    if negative:
        dem = -dem

    for dir in dirs:  # noqa: A001
        opn[dir, :, :] = _openness_dir(dem, res, L, dir)

    if mean:
        opn = opn.mean(axis=0)

    return opn


@njit(cache=True, parallel=True)
def _openness_dir(dem, res, L, dir):  # noqa: A002
    """
    Calculate topographic openness for a DEM and a single compass direction.

    Parameters
    ----------
    dem : ndarray
        Terrain elevation (m).

    res : float
        DEM resolution (m).

    L : float
        Radial distance to consider for calculating openness for each pixel (m).

    dir : int
        Direction for which to calculate openness, ranging from 0 (northwest) to
        7 (west).

    Returns
    -------
    opn_dir : ndarray
        Openness (radians).
    """
    opn_dir = np.full(dem.shape, np.inf)

    for i in prange(int(np.ceil(L / res))):
        dist = res * (i + 1) * [1, np.sqrt(2)][dir % 2]
        Z_shift = _shift_arr_retain(dem, dir, i)
        angle = np.pi / 2 - np.arctan2(Z_shift - dem, dist)

        idxs = np.flatnonzero(angle < opn_dir)
        opn_dir.ravel()[idxs] = angle.ravel()[idxs]

    return opn_dir


def _shift_arr(M, dir, n, mode="retain"):  # noqa: A002
    """
    Shift an array along one of the eight (inter)cardinal directions.

    Parameters
    ----------
    M : ndarray
        Input array.

    dir : int
        Direction along to shift the array, ranging from 0 (north) to 7
        (northwest).

    n : int
        Number of pixels to be shifted.

    mode : str, default 'retain'
        If 'retain', pixels are padded with the values from the original array.
        All other values (e.g., 'edge' for padding with the edge values) are
        passed to np.pad().

    Returns
    -------
    S : ndarray
        Shifted array.
    """
    if mode == "retain":
        return _shift_arr_retain(M, dir, n)
    else:
        if dir == 0:  # north
            return np.pad(M, ((0, n), (0, 0)), mode=mode)[n:, :]
        elif dir == 1:  # northeast
            return np.pad(M, ((0, n), (n, 0)), mode=mode)[n:, :-n]
        elif dir == 2:  # east
            return np.pad(M, ((0, 0), (n, 0)), mode=mode)[:, :-n]
        elif dir == 3:  # southeast
            return np.pad(M, ((n, 0), (n, 0)), mode=mode)[:-n, :-n]
        elif dir == 4:  # south
            return np.pad(M, ((n, 0), (0, 0)), mode=mode)[:-n, :]
        elif dir == 5:  # southwest
            return np.pad(M, ((n, 0), (0, n)), mode=mode)[:-n, n:]
        elif dir == 6:  # west
            return np.pad(M, ((0, 0), (0, n)), mode=mode)[:, n:]
        elif dir == 7:  # northwest
            return np.pad(M, ((0, n), (0, n)), mode=mode)[n:, n:]


@njit(cache=True)
def _shift_arr_retain(M, dir, n):  # noqa: A002
    """
    Shift an array along one of the eight (inter)cardinal directions.
    Pixels padded to the edges of the axes retain the value from the
    original array.

    Parameters
    ----------
    M : ndarray
        Input array.

    dir : int
        Direction along to shift the array, ranging from 0 (north) to 7
        (northwest).

    n : int
        Number of pixels to be shifted.

    Returns
    -------
    S : ndarray
        Shifted array.
    """
    S = M.copy()

    if dir == 0:  # north
        S[: -n - 1, :] = M[1 + n :, :]
    elif dir == 1:  # northeast
        S[: -n - 1, 1 + n :] = M[1 + n :, : -n - 1]
    elif dir == 2:  # east
        S[:, 1 + n :] = M[:, : -n - 1]
    elif dir == 3:  # southeast
        S[1 + n :, 1 + n :] = M[: -n - 1, : -n - 1]
    elif dir == 4:  # south
        S[1 + n :, :] = M[: -n - 1, :]
    elif dir == 5:  # southwest
        S[1 + n :, : -n - 1] = M[: -n - 1, 1 + n :]
    elif dir == 6:  # west
        S[:, : -n - 1] = M[:, 1 + n :]
    elif dir == 7:  # northwest
        S[: -n - 1, : -n - 1] = M[1 + n :, 1 + n :]

    return S



from .init import initialize
from .soil import (
    soil_heat_flux,
    soil_properties,
    soil_temperature,
)

__all__ = [
    "initialize",
    "soil_heat_flux",
    "soil_properties",
    "soil_temperature",
]



from .atmosphere import (
    absolute_humidity,
    atmospheric_pressure,
    clear_sky_emissivity,
    cloud_factor_from_cloud_fraction,
    cloud_fraction_from_cloud_factor,
    cloud_fraction_from_humidity,
    dew_point_temperature,
    dry_air_density,
    latent_heat_of_vaporization,
    log_wind_profile,
    precipitable_water,
    precipitation_phase,
    pressure_to_altitude,
    psychrometric_constant,
    relative_humidity,
    saturation_vapor_pressure,
    specific_heat_capacity_moist_air,
    specific_humidity,
    vapor_pressure,
    wet_bulb_temperature,
    wind_from_uv,
    wind_to_uv,
)
from .interpolation import interpolate_param, interpolate_station_data
from .precipcorr import correct_station_precipitation

__all__ = [
    "absolute_humidity",
    "atmospheric_pressure",
    "clear_sky_emissivity",
    "cloud_factor_from_cloud_fraction",
    "cloud_fraction_from_cloud_factor",
    "cloud_fraction_from_humidity",
    "correct_station_precipitation",
    "dew_point_temperature",
    "dry_air_density",
    "interpolate_param",
    "interpolate_station_data",
    "latent_heat_of_vaporization",
    "log_wind_profile",
    "precipitable_water",
    "precipitation_phase",
    "pressure_to_altitude",
    "psychrometric_constant",
    "relative_humidity",
    "saturation_vapor_pressure",
    "specific_heat_capacity_moist_air",
    "specific_humidity",
    "vapor_pressure",
    "wet_bulb_temperature",
    "wind_from_uv",
    "wind_to_uv",
]



import numpy as np
from numba import njit

from openamundsen import tridiag


@njit(cache=True)
def temp_change(
    dx,
    dt,
    T,
    therm_cond,
    T_bottom,
    dx_bottom,
    therm_cond_bottom,
    top_heat_flux,
    heat_cap,
    replace_nan=True,
):
    """
    Calculate the change in temperature over time by implicitly solving the
    one-dimensional heat equation following [1].

    Parameters
    ----------
    dx : ndarray
        Layer thicknesses (m).

    dt : float
        Timestep (s).

    T : ndarray
        Current temperatures (K).

    therm_cond : ndarray
        Thermal conductivities (W m-1 K-1).

    T_bottom : float
        Temperature of the bottom layer (K).

    dx_bottom : float
        Thickness of the bottom layer (m).

    top_heat_flux : float
        Heat flux from the top (W m-2).

    heat_cap : ndarray
        Areal heat capacities (J K-1 m-2).

    replace_nan : bool, default True
        Replace nan values in the calculated temperature changes with 0.

    Returns
    -------
    dT : ndarray
        Change in temperature over the timestep.

    References
    ----------
    .. [1] Essery, R. (2015). A factorial snowpack model (FSM 1.0).
       Geoscientific Model Development, 8(12), 3867–3876.
       https://doi.org/10.5194/gmd-8-3867-2015
    """
    N = len(T)  # number of layers

    # Calculate thermal conductivity between layers
    therm_cond_between = np.full(N, np.nan)
    for k in range(N - 1):
        therm_cond_between[k] = 2 / (dx[k] / therm_cond[k] + dx[k + 1] / therm_cond[k + 1])
    therm_cond_between[N - 1] = 2 / (dx[N - 1] / therm_cond[N - 1] + dx_bottom / therm_cond_bottom)

    if N == 1:
        temp_change = np.array(
            [
                (top_heat_flux + therm_cond_between[0] * (T_bottom - T[0]))
                * dt
                / (heat_cap[0] + therm_cond_between[0] * dt)
            ]
        )
    else:
        a = np.zeros(N)  # below-diagonal matrix elements
        b = np.zeros(N)  # diagonal matrix elements
        c = np.zeros(N)  # above-diagonal matrix elements
        d = np.zeros(N)  # right hand side of the matrix equation

        a[0] = 0
        b[0] = heat_cap[0] + therm_cond_between[0] * dt
        c[0] = -therm_cond_between[0] * dt
        d[0] = (top_heat_flux - therm_cond_between[0] * (T[0] - T[1])) * dt

        for k in range(1, N - 1):
            a[k] = c[k - 1]
            b[k] = heat_cap[k] + (therm_cond_between[k - 1] + therm_cond_between[k]) * dt
            c[k] = -therm_cond_between[k] * dt
            d[k] = (
                therm_cond_between[k - 1] * (T[k - 1] - T[k]) * dt
                + therm_cond_between[k] * (T[k + 1] - T[k]) * dt
            )

        k = N - 1
        a[k] = c[k - 1]
        b[k] = heat_cap[k] + (therm_cond_between[k - 1] + therm_cond_between[k]) * dt
        c[k] = 0
        d[k] = (
            therm_cond_between[k - 1] * (T[k - 1] - T[k]) * dt
            + therm_cond_between[k] * (T_bottom - T[k]) * dt
        )

        temp_change = tridiag.solve_tridiag(a, b, c, d)

    if replace_nan:
        temp_change[np.isnan(temp_change)] = 0.0

    return temp_change


def temp_change_array(
    dx,
    dt,
    T,
    therm_cond,
    T_bottom,
    dx_bottom,
    therm_cond_bottom,
    top_heat_flux,
    heat_cap,
    replace_nan=True,
):
    """
    Array version of temp_change().
    Can be applied if the number of layers is the same for all pixels.

    Parameters
    ----------
    dx : ndarray
        Layer thicknesses (m).

    dt : float
        Timestep (s).

    T : ndarray(ndim=2)
        Current temperatures (K).

    therm_cond : ndarray(ndim=2)
        Thermal conductivities (W m-1 K-1).

    T_bottom : ndarray(ndim=1)
        Temperature of the bottom layer (K).

    dx_bottom : ndarray(ndim=1)
        Thickness of the bottom layer (m).

    top_heat_flux : ndarray(ndim=1)
        Heat flux from the top (W m-2).

    heat_cap : ndarray(ndim=2)
        Areal heat capacities (J K-1 m-2).

    replace_nan : bool, default True
        Replace nan values in the calculated temperature changes with 0.

    Returns
    -------
    dT : ndarray(ndim=2)
        Change in temperature over the timestep.
    """
    N = T.shape[0]  # number of layers

    # Calculate thermal conductivity between layers
    therm_cond_between = np.full(T.shape, np.nan)
    for k in range(N - 1):
        therm_cond_between[k, :] = 2 / (
            dx[k, :] / therm_cond[k, :] + dx[k + 1, :] / therm_cond[k + 1, :]
        )
    therm_cond_between[N - 1, :] = 2 / (
        dx[N - 1, :] / therm_cond[N - 1, :] + dx_bottom / therm_cond_bottom
    )

    if N == 1:
        temp_change = np.atleast_2d(
            (top_heat_flux + therm_cond_between[0, :] * (T_bottom - T[0, :]))
            * dt
            / (heat_cap[0, :] + therm_cond_between[0, :] * dt)
        )
    else:
        a = np.zeros(T.shape)  # below-diagonal matrix elements
        b = np.zeros(T.shape)  # diagonal matrix elements
        c = np.zeros(T.shape)  # above-diagonal matrix elements
        d = np.zeros(T.shape)  # right hand side of the matrix equation

        a[0, :] = 0
        b[0, :] = heat_cap[0, :] + therm_cond_between[0, :] * dt
        c[0, :] = -therm_cond_between[0, :] * dt
        d[0, :] = (top_heat_flux - therm_cond_between[0, :] * (T[0, :] - T[1, :])) * dt

        for k in range(1, N - 1):
            a[k, :] = c[k - 1, :]
            b[k, :] = (
                heat_cap[k, :] + (therm_cond_between[k - 1, :] + therm_cond_between[k, :]) * dt
            )
            c[k, :] = -therm_cond_between[k, :] * dt
            d[k, :] = (
                therm_cond_between[k - 1, :] * (T[k - 1, :] - T[k, :]) * dt
                + therm_cond_between[k, :] * (T[k + 1, :] - T[k, :]) * dt
            )

        k = N - 1
        a[k, :] = c[k - 1, :]
        b[k, :] = heat_cap[k, :] + (therm_cond_between[k - 1, :] + therm_cond_between[k, :]) * dt
        c[k, :] = 0
        d[k, :] = (
            therm_cond_between[k - 1, :] * (T[k - 1, :] - T[k, :]) * dt
            + therm_cond_between[k, :] * (T_bottom - T[k, :]) * dt
        )

        temp_change = tridiag.solve_tridiag_array(a, b, c, d)

    if replace_nan:
        temp_change[np.isnan(temp_change)] = 0.0

    return temp_change



from .irradiance import (
    clear_sky_shortwave_irradiance,
    longwave_irradiance,
    shortwave_irradiance,
)
from .shadows import shadows
from .sunparams import (
    day_angle,
    declination_angle,
    equation_of_time,
    hour_angle,
    sun_parameters,
    sun_vector,
)

__all__ = [
    "clear_sky_shortwave_irradiance",
    "day_angle",
    "declination_angle",
    "equation_of_time",
    "hour_angle",
    "longwave_irradiance",
    "shadows",
    "shortwave_irradiance",
    "sun_parameters",
    "sun_vector",
]



import pprint
from dataclasses import dataclass

import numpy as np
import pandas as pd
import rasterio
import xarray as xr
from loguru import logger

from openamundsen import constants, errors, util


@dataclass
class PointOutputVariable:
    """
    Class for defining a point output variable.

    Parameters
    ----------
    var_name : str
        Name of the state variable (e.g. "meteo.temp").

    output_name : str, optional
        Output name (by default the variable name, e.g. "temp" for the variable
        "meteo.temp")
    """

    var_name: str
    output_name: str = None

    def __post_init__(self):
        if self.output_name is None:
            self.output_name = self.var_name.split(".")[-1]


@dataclass
class OutputPoint:
    """
    Class for defining an output point.

    Parameters
    ----------
    name : str
        Point name.

    lon : float
        Longitude (degrees).

    lat : float
        Latitude (degrees).

    alt : float
        Altitude (m).

    x : float
        x coordinate in the model CRS (m).

    y : float
        y coordinate in the model CRS (m).

    col : int
        Column index of the point within the model grid.

    row : int
        Row index of the point within the model grid.
    """

    name: str
    lon: float
    lat: float
    alt: float
    x: float
    y: float
    col: int
    row: int


_default_output_vars = [
    PointOutputVariable("meteo.temp"),
    PointOutputVariable("meteo.precip"),
    PointOutputVariable("meteo.snowfall"),
    PointOutputVariable("meteo.rainfall"),
    PointOutputVariable("meteo.rel_hum"),
    PointOutputVariable("meteo.wind_speed"),
    PointOutputVariable("meteo.sw_in"),
    PointOutputVariable("meteo.sw_out"),
    PointOutputVariable("meteo.lw_in"),
    PointOutputVariable("meteo.lw_out"),
    PointOutputVariable("meteo.sw_in_clearsky"),
    PointOutputVariable("meteo.dir_in_clearsky"),
    PointOutputVariable("meteo.diff_in_clearsky"),
    PointOutputVariable("meteo.cloud_factor"),
    PointOutputVariable("meteo.cloud_fraction"),
    PointOutputVariable("meteo.wet_bulb_temp"),
    PointOutputVariable("meteo.dew_point_temp"),
    PointOutputVariable("meteo.atmos_press"),
    PointOutputVariable("meteo.sat_vap_press"),
    PointOutputVariable("meteo.vap_press"),
    PointOutputVariable("meteo.spec_hum"),
    #
    PointOutputVariable("surface.temp", "surface_temp"),
    PointOutputVariable("surface.heat_flux", "surface_heat_flux"),
    PointOutputVariable("surface.sens_heat_flux"),
    PointOutputVariable("surface.lat_heat_flux"),
    PointOutputVariable("surface.advective_heat_flux"),
    PointOutputVariable("surface.albedo", "surface_albedo"),
    #
    PointOutputVariable("soil.temp", "soil_temp"),
    PointOutputVariable("soil.heat_flux", "soil_heat_flux"),
    #
    PointOutputVariable("snow.swe", "swe"),
    PointOutputVariable("snow.depth", "snow_depth"),
    PointOutputVariable("snow.temp", "snow_temp"),
    PointOutputVariable("snow.thickness", "snow_thickness"),
    PointOutputVariable("snow.density", "snow_density"),
    PointOutputVariable("snow.ice_content"),
    PointOutputVariable("snow.liquid_water_content"),
    PointOutputVariable("snow.melt"),
    PointOutputVariable("snow.runoff"),
    PointOutputVariable("snow.sublimation"),
    PointOutputVariable("snow.refreezing"),
]


class PointOutputManager:
    """
    Class for managing and storing point output data (i.e. time series for
    selected point locations). The data is stored in an xarray Dataset and
    written to file at specified intervals.

    Parameters
    ----------
    model : OpenAmundsen
        openAMUNDSEN model instance.
    """

    def __init__(self, model):
        vars = []  # noqa: A001
        points = []
        config = model.config.output_data.timeseries

        # Initialize write dates
        if config.format == "memory":
            write_dates = []
        else:
            write_dates = pd.date_range(
                start=model.config.start_date,
                end=model.config.end_date,
                freq=util.to_offset(config.write_freq),
            )
            write_dates = write_dates[
                (write_dates >= model.config.start_date) & (write_dates <= model.config.end_date)
            ]

            # Add last timestep of the model run if it is not included anyway
            if model.config.end_date not in write_dates:
                write_dates = write_dates.append(pd.DatetimeIndex([model.config.end_date]))
        self.write_dates = write_dates

        # Add default output variables
        if config.add_default_variables:
            for var in _default_output_vars:
                vars.append(var)

        # Add user defined output variables
        for var_cfg in config.variables:
            vars.append(
                PointOutputVariable(
                    var_name=var_cfg["var"],
                    output_name=var_cfg["name"] if "name" in var_cfg else None,
                )
            )

        # Check if all variable specifications are valid
        for var in vars:
            try:
                _ = model.state[var.var_name]
            except (AttributeError, KeyError) as err:
                raise errors.ConfigurationError(
                    f"Invalid time series output variable: {var.var_name}"
                ) from err

        # Check if there are any duplicate output names
        if len({v.output_name for v in vars}) < len(vars):
            raise errors.ConfigurationError(
                "Duplicate output names in time series output configuration.\n"
                f"List of variables:\n{pprint.pformat(vars)}"
            )

        # Add default output points (= stations within ROI)
        if config.add_default_points:
            stations = model.meteo.sel(station=model.meteo.within_roi)
            df = stations[["lon", "lat", "alt", "x", "y", "col", "row"]].to_dataframe()

            for point in df.itertuples():
                points.append(
                    OutputPoint(
                        name=point.Index,
                        lon=point.lon,
                        lat=point.lat,
                        alt=model.state.base.dem[point.row, point.col],
                        x=point.x,
                        y=point.y,
                        col=point.col,
                        row=point.row,
                    )
                )

        # Add additional output points
        for point_num, point in enumerate(config.points):
            if "name" in point:
                point_name = point["name"]
            else:
                point_name = f"point{point_num + 1}"

            # Check if point name is already in use
            if point_name in [p.name for p in points]:
                raise errors.ConfigurationError(f"Duplicate point name: {point_name}")

            lon, lat = util.transform_coords(
                point["x"],
                point["y"],
                model.config.crs,
                constants.CRS_WGS84,
            )

            row, col = rasterio.transform.rowcol(
                model.grid.transform,
                point["x"],
                point["y"],
            )
            row = int(row)
            col = int(col)

            # Check if point is within the grid boundaries
            if row < 0 or row >= model.grid.rows or col < 0 or col >= model.grid.cols:
                raise errors.ConfigurationError(
                    f"Output point is outside of the grid boundaries: {dict(point)}"
                )

            # Check if point is within the ROI (if not, still allow it but raise a warning)
            if not model.grid.roi[row, col]:
                logger.warning(f"Output point is outside of the ROI: {dict(point)}")

            points.append(
                OutputPoint(
                    name=point_name,
                    lon=lon,
                    lat=lat,
                    alt=model.state.base.dem[row, col],
                    x=point["x"],
                    y=point["y"],
                    col=col,
                    row=row,
                )
            )

        self.model = model
        self.vars = vars
        self.points = points
        self.point_cols = [p.col for p in points]
        self.point_rows = [p.row for p in points]
        self.data = None
        self.format = config.format

    def _create_dataset(self, dates):
        """
        Create a Dataset covering the specified dates and all point locations
        and output variables according to the model run configuration.

        Parameters
        ----------
        dates : list of datetime-like

        Returns
        -------
        ds : xr.Dataset
        """
        data = {}
        three_dim_coords = {}
        for var in self.vars:
            meta = self.model.state.meta(var.var_name)
            attrs = {}

            if np.issubdtype(meta.dtype, np.integer):
                dtype = np.int32
                fill_value = np.iinfo(dtype).min
            else:
                dtype = np.float32
                fill_value = np.nan

            attrs["_FillValue"] = fill_value

            for attr in ("standard_name", "long_name", "units"):
                attr_val = getattr(meta, attr)
                if attr_val is not None:
                    attrs[attr] = attr_val

            if meta.dim3 == 0:  # 2-dimensional variable
                var_def = (
                    ["time", "point"],
                    np.full((len(dates), len(self.points)), fill_value, dtype=dtype),
                    attrs,
                )
            else:  # 3-dimensional variable
                category = self.model.state.parse(var.var_name)[0]
                coord_name = f"{category}_layer"

                if category in three_dim_coords:
                    if three_dim_coords[coord_name] != meta.dim3:
                        # We assume that all 3-dimensional variables within a category have the
                        # same shape (e.g. "soil.temp" must have the same shape as
                        # "soil.therm_cond"); varying numbers of layers within a category are not
                        # supported
                        raise ValueError("Inconsistent length of third variable dimension")
                else:
                    three_dim_coords[coord_name] = meta.dim3

                var_def = (
                    ["time", coord_name, "point"],
                    np.full((len(dates), meta.dim3, len(self.points)), fill_value, dtype=dtype),
                    attrs,
                )

            data[var.output_name] = var_def

        coords = {
            "time": (["time"], dates),
            "point": (["point"], [p.name for p in self.points]),
            "lon": (["point"], [p.lon for p in self.points]),
            "lat": (["point"], [p.lat for p in self.points]),
            "alt": (["point"], [p.alt for p in self.points]),
            "x": (["point"], [p.x for p in self.points]),
            "y": (["point"], [p.y for p in self.points]),
        }

        # Add 3-dimensional coordinates
        for coord_name, coord_len in three_dim_coords.items():
            coords[coord_name] = ([coord_name], np.arange(coord_len))

        return xr.Dataset(data, coords=coords)

    def _current_chunk_dates(self):
        """
        Return the dates of the current "chunk", i.e., the period between the
        last date for which the output time series were written and the next
        date for which they will be written.

        Returns
        -------
        dates : pd.DatetimeIndex
        """
        if self.format == "memory":
            dates = self.model.dates
        else:
            idxs = np.flatnonzero(self.write_dates >= self.model.date)
            next_write_date = self.write_dates[idxs[0]]

            idxs = np.flatnonzero(self.write_dates < self.model.date)
            if len(idxs) == 0:
                dates = self.model.dates[self.model.dates <= next_write_date]
            else:
                prev_write_date = self.write_dates[idxs[-1]]
                dates = self.model.dates[
                    (self.model.dates > prev_write_date) & (self.model.dates <= next_write_date)
                ]

        return dates

    def update(self):
        """
        Update the point output data for the current time step, i.e., write the
        output variable values for all point locations to the internal dataset.
        """
        logger.debug("Updating point outputs")

        if self.data is None:
            self.data = self._create_dataset(self._current_chunk_dates())

        ds = self.data
        date = self.model.date

        # Get the index for the current date because directly writing to the
        # underlying numpy arrays of the xarray variables
        # (ds_var.values[date_idx, :]) is a lot faster than using xarray with
        # label-based indexing (ds_var.loc[date, :]).
        # Using np.argmax is equivalent to using np.where to get the index,
        # but is faster because it stops at the first match.
        date_idx = np.argmax(ds.time.values == date.to_datetime64())

        # Update dataset
        for var in self.vars:
            var_data = self.model.state[var.var_name]
            ds_var = ds.variables[var.output_name]  # faster than ds[var.output_name]

            if var_data.ndim == 2:
                ds_var.values[date_idx, :] = var_data[self.point_rows, self.point_cols]
            else:  # 3-dimensional variable
                ds_var.values[date_idx, :, :] = var_data[:, self.point_rows, self.point_cols]

        # Write data to file
        # If we are at the first write date, simple write the file (i.e. overwrite possibly
        # existing files), otherwise merge the already written dataset with the in-memory one.
        if self.format != "memory" and date == ds.indexes["time"][-1]:
            logger.debug("Writing point outputs")

            if self.format == "netcdf":
                filename = self.model.config.results_dir / "output_timeseries.nc"

                if date == self.write_dates[0]:
                    ds.to_netcdf(filename)
                else:
                    with xr.open_dataset(filename) as old_ds:
                        # Handle an issue introduced in xarray v2022.06.0 - without this line, the
                        # time index somehow loses the time information and keeps only the date part
                        # (triggered by test_point_output.py::test_write_freq)
                        old_ds["time"] = old_ds.indexes["time"]

                        ds_merge = xr.concat([old_ds, ds], "time")

                    ds_merge.to_netcdf(filename)
            elif self.format == "csv":
                ds_out = ds.copy()

                # If there are 3-dimensional variables, convert them to 2-dimensional variables
                # before writing the CSV files
                # (e.g. a 3-dimensional variable "soil_temp" with 4 layers is converted to 4
                # variables "soil_temp0", "soil_temp1", "soil_temp2", "soil_temp3")
                for var in self.vars:
                    meta = self.model.state.meta(var.var_name)

                    if meta.dim3 > 0:
                        for i in range(meta.dim3):
                            ds_out[f"{var.output_name}{i}"] = ds[var.output_name].loc[:, i, :]

                        ds_out = ds_out.drop_vars(var.output_name)

                for point in self.points:
                    filename = self.model.config.results_dir / f"point_{point.name}.csv"

                    ds_out_point = ds_out.sel(point=point.name)

                    # Drop all coordinate variables except "time" (i.e. "lon", "lat", etc.)
                    # so that they are not in the resulting dataframe when calling to_dataframe()
                    ds_out_point = ds_out_point.drop_vars(list(set(ds_out_point.coords) - {"time"}))

                    df = ds_out_point.to_dataframe()

                    if date == self.write_dates[0]:
                        df.to_csv(filename)
                    else:
                        df.to_csv(filename, mode="a", header=False)

            self.data = None



import rasterio

from openamundsen.errors import RasterFileError


def read_raster_metadata(filename, crs=None):
    """
    Return metadata for a raster file.

    Parameters
    ----------
    filename : str or pathlib.Path

    crs : str, default None
        CRS to be set in the returned dict if the raster file does not contain
        CRS information. Must be a string parsable by rasterio.crs.CRS.from_string
        (e.g. "epsg:32632").

    Returns
    -------
    meta : dict
        Dictionary containing the following keys:
        - 'rows' (number of rows),
        - 'cols' (number of columns)
        - 'shape' ((rows, cols) tuple)
        - 'resolution' (grid resolution)
        - 'crs': coordinate reference system
        - 'transform' (georeferencing transformation parameters)
    """
    meta = {}

    with rasterio.open(filename) as ds:
        if not (ds.res[0] == ds.res[1] == abs(ds.res[0])):
            raise RasterFileError("Raster file must have equal x and y resolution")

        meta["rows"] = ds.meta["height"]
        meta["cols"] = ds.meta["width"]
        meta["shape"] = (meta["rows"], meta["cols"])
        meta["resolution"] = ds.res[0]
        meta["crs"] = ds.crs
        meta["transform"] = ds.meta["transform"]

        if meta["crs"] is None and crs is not None:
            meta["crs"] = rasterio.crs.CRS.from_string(crs)

    return meta


def read_raster_file(filename, check_meta=None, fill_value=None, dtype=None):
    """
    Read a raster file.

    Parameters
    ----------
    filename : str or pathlib.Path

    check_meta : dict, default None
        A metadata dictionary (as returned by `read_raster_metadata` to compare
        the current raster metadata with. If the metadata of the two rasters
        does not match (e.g., if the number of rows or columns differs) a
        RasterFileError is raised.

    fill_value : numeric, default None
        Value to use for nodata pixels in the source file. If None, nodata
        values are left unchanged.

    dtype : dtype, default None
        Data type to cast the data array to.

    Returns
    -------
    data : np.ndarray
    """
    if check_meta is not None:
        # compare only rows, cols, resolution and transform and not additional attributes
        # possibly stored in the check_meta object (such as x and y coordinates, CRS, etc.)
        cmp_keys = [
            "rows",
            "cols",
            "resolution",
            "transform",
        ]

        meta = read_raster_metadata(filename)

        d1 = {k: meta[k] for k in cmp_keys}
        d2 = {k: check_meta[k] for k in cmp_keys}

        if d1 != d2:
            raise RasterFileError(f"Metadata mismatch for {filename}")

    with rasterio.open(filename) as ds:
        data = ds.read(1, masked=(fill_value is not None))

        if dtype is not None:
            data = data.astype(dtype)

        if fill_value is not None:
            data = data.filled(fill_value)

    return data


def write_raster_file(filename, data, transform, **kwargs):
    """
    Write a raster file.

    Parameters
    ----------
    filename : str or pathlib.Path

    data : ndarray
        Array to be written.

    transform : rasterio.Affine
        Georeferencing transformation parameters.

    **kwargs
        Additional keyword arguments to be passed over to `rasterio.open`.
    """
    meta = {
        "driver": "AAIGrid",
        "dtype": data.dtype,
        "nodata": None,
        "width": data.shape[1],
        "height": data.shape[0],
        "count": 1,
        "crs": None,
        "transform": transform,
    }
    meta.update(kwargs)

    with rasterio.open(filename, "w", **meta) as ds:
        ds.write(data, 1)
